import logging
import os
import os.path
import yaml
import stat

from packy_agent.utils.misc import atomic_write


logger = logging.getLogger(__name__)


class BaseConfiguration(object):

    PACKY_CONFIGURATION_KEY = 'packy'

    def __init__(self):
        self.common_configuration = None
        self.local_configuration = {}
        self.local_configuration_mtime = None

    def get_local_configuration_file_path(self):
        raise NotImplementedError('Method must be implemented by child classes')

    def load_local_configuration(self, content=None, force=False):
        if content is not None:
            self.local_configuration = yaml.load(content)
            return True

        file_path = self.get_local_configuration_file_path()
        if not os.path.isfile(file_path):
            logger.debug('Configuration file %s does not exist', file_path)
            self.local_configuration.clear()
            self.local_configuration_mtime = None
            return True

        # TODO(dmu) LOW: Because we write to configuration file during run time it potentially
        #                may disappear at this point (if we delete it).
        #                Consider handling this situation
        mtime = os.path.getmtime(file_path)
        if mtime == self.local_configuration_mtime and not force:
            return False

        with open(file_path) as f:
            self.local_configuration = yaml.load(f)

        self.local_configuration_mtime = mtime
        return True

    def save_local_configuration(self, content=None):
        if content is None:
            config_yaml = yaml.safe_dump(dict(self.local_configuration), default_flow_style=False)
        else:
            if isinstance(content, basestring):
                yaml.load(content)  # for basic validation
                config_yaml = content
            else:
                config_yaml = yaml.safe_dump(content, default_flow_style=False)

        local_configuration_file_path = self.get_local_configuration_file_path()
        dir_path = os.path.dirname(local_configuration_file_path)
        if not os.path.isdir(dir_path):
            raise Exception('Directory for configuration file does not exist: {}'.format(dir_path))

        logger.debug('Saving local configuration to %s', local_configuration_file_path)
        with atomic_write(local_configuration_file_path, overwrite=True) as f:
            f.write(config_yaml)

        # TODO(dmu) MEDIUM: Consider not doing chmod, but it is convenient for development now
        os.chmod(local_configuration_file_path,
                 stat.S_IRUSR | stat.S_IWUSR | stat.S_IRGRP | stat.S_IWGRP | stat.S_IROTH |
                 stat.S_IWOTH)

    def is_debug_mode(self):
        return os.getenv('PACKY_AGENT_DEBUG', '').lower() in ('1', 'y', 'yes', 'true')



from packy_agent.tasks.management.heartbeat import *
from packy_agent.tasks.management.restart import *
from packy_agent.tasks.management.update_config import *
from packy_agent.tasks.management.install_and_upgrade import *



from __future__ import absolute_import

import speedtest


def run_speedtest(server_id):
    speedtest_client = speedtest.Speedtest()
    speedtest_client.get_servers([server_id])
    speedtest_client.get_best_server()
    speedtest_client.download()
    speedtest_client.upload()

    return speedtest_client.results



from __future__ import absolute_import

import logging
import struct
import time

from ryu.lib.packet.icmp import icmp, echo
from ryu.lib.packet.packet import Packet

from packy_agent.utils.misc import generate_random_string

ICMP_PROTOCOL = 'icmp'

sequence = {
    ICMP_PROTOCOL: 1
}
logger = logging.getLogger(__name__)


def get_seq_no(proto):
    rv = sequence[proto]
    sequence[proto] += 1
    sequence[proto] &= 0xFFFF
    return rv


TIME_FORMAT = '<II'
TIME_SIZE_BYTES = struct.calcsize(TIME_FORMAT)


def encode_timestamp(timestamp):
    sent_time_seconds = int(timestamp)
    sent_time_microseconds = int((timestamp - sent_time_seconds) * 1000000)
    return struct.pack(TIME_FORMAT, sent_time_seconds, sent_time_microseconds)


def decode_timestamp(timestamp):
    sent_time_seconds, sent_time_microseconds = struct.unpack(TIME_FORMAT, timestamp)
    return sent_time_seconds + sent_time_microseconds / 1000000.


def send_icmp_echo_request(raw_socket, destination_ip_address, id_, seq_no, payload_size_bytes=56,
                           payload=None):
    sent_time = time.time()
    time_part = encode_timestamp(sent_time)
    if payload is None:
        if payload_size_bytes < 0:
            raise ValueError('payload_size_bytes must be greater or equal to 0')

        if payload_size_bytes < TIME_SIZE_BYTES:
            payload = generate_random_string(payload_size_bytes)
        else:
            payload = time_part + generate_random_string(payload_size_bytes - TIME_SIZE_BYTES)
    else:
        payload = time_part + payload

    packet = Packet()
    packet.add_protocol(icmp(data=echo(id_, seq_no, payload)))
    packet.serialize()

    raw_socket.sendto(packet.data, (destination_ip_address, 1))  # 1 - is a dummy port for ICMP
    logger.debug('Sent ICMP echo request({}) packet to {} with id:{} and seq:{}'.format(
        packet.protocols[0].type, destination_ip_address, id_, seq_no))
    return sent_time, payload



import os
import logging
import yaml

from kombu import Queue, Exchange
from kombu.common import Broadcast

from packy_agent.utils.collections import deep_update

from packy_agent.configuration import common as common_configuration
from packy_agent.configuration.base import BaseConfiguration
from packy_agent.exceptions import ImproperlyConfigured
from packy_agent.utils.pkg_resources import get_package_file_content
from packy_agent.utils.network import get_hostname_from_url
from packy_agent.clients.packy_server import packy_server_client_factory
from packy_agent.configuration.control_server.base import configuration as cs_configuration


logger = logging.getLogger(__name__)


class PackyAgentConfiguration(BaseConfiguration):

    CELERY_CONFIGURATION_KEY = 'celery'
    AGENT_CONFIGURATION_KEY = 'agent'

    def __init__(self):
        super(PackyAgentConfiguration, self).__init__()

        self.celery_configuration = None
        self.packy_agent_configuration = None

    def update_local_configuration_from_server(self):
        logger.info('Updating configuration file...')
        agent_id = configuration.get_agent_id()
        if agent_id is None:
            logger.error('Could not update config, because Packy Agent ID (packy.agent.agent_id) '
                         'is not known')
            return

        client = packy_server_client_factory.get_client_by_configuration(configuration)
        response = client.get_agent_config(agent_id)

        if response.status_code == 200:
            cs_server_base_url = cs_configuration.get_server_base_url()
            self.save_local_configuration(content=response.content)

            # Update Packy Server base URL if it changed
            server_base_url = self.get_server_base_url()
            if server_base_url != cs_server_base_url:
                cs_configuration.local_configuration.setdefault('packy', {}).setdefault(
                    'control_server', {})['server_base_url'] = server_base_url
                cs_configuration.save_local_configuration()
            # TODO(dmu) HIGH: If Celery related configuration is updated, for example `broker_url`
            #                 then Celery should be restarted
        else:
            logger.error('Error while updating config: HTTP%s %s',
                         response.status_code, response.content)

    def get_local_configuration_file_path(self):
        return (os.getenv(common_configuration.AGENT_CONFIG_ENVVAR_NAME) or
                common_configuration.DEFAULT_AGENT_CONFIG_PATH)

    def get_celery_configuration(self):
        was_updated = self.load_local_configuration()
        if was_updated or self.celery_configuration is None:
            agent_id = self.get_agent_id()
            if agent_id is None:
                raise ImproperlyConfigured('Packy Agent id (packy.agent.agent_id) is not set,'
                                           'unable to get celery configuration')

            celery_configuration = yaml.load(
                get_package_file_content('packy_agent.configuration.agent', 'celery.yaml'))
            deep_update(celery_configuration,
                        self.local_configuration[self.CELERY_CONFIGURATION_KEY])

            # TODO(dmu) HIGH: This is for compatibility reasons, remove it once Packy Server is
            #                 upgraded to v0.0.13
            if self.local_configuration[self.CELERY_CONFIGURATION_KEY].get('broker_url') is None:
                broker_url = self.local_configuration.get('broker_url')
                if broker_url is not None:
                    celery_configuration['broker_url'] = broker_url

            str_agent_id = str(agent_id)
            celery_configuration['task_default_queue'] = str_agent_id
            celery_configuration['task_default_routing_key'] = str_agent_id
            celery_configuration['task_queues'] = (
                Queue(str_agent_id,
                      Exchange(celery_configuration['task_default_exchange'], delivery_mode=1),
                      routing_key=str_agent_id, auto_delete=True, durable=False,
                      queue_arguments={'x-message-ttl': 10000}),
                Broadcast('broadcast_tasks', auto_delete=True, durable=False, exclusive=True,
                          delivery_mode=1,
                          queue_arguments={'x-message-ttl': 10000, 'x-expires': 10000}),
            )

            self.celery_configuration = celery_configuration

        return self.celery_configuration

    def get_packy_agent_configuration(self):
        was_updated = self.load_local_configuration()
        if was_updated or self.packy_agent_configuration is None:
            packy_agent_configuration = yaml.load(
                get_package_file_content('packy_agent.configuration.agent', 'custom.yaml'))
            deep_update(packy_agent_configuration,
                        (self.local_configuration.get(self.PACKY_CONFIGURATION_KEY) or {}).get(
                         self.AGENT_CONFIGURATION_KEY) or {})

            self.packy_agent_configuration = packy_agent_configuration

        return self.packy_agent_configuration

    def get_watchdog_configuration(self):
        return self.get_packy_agent_configuration()['watchdog']

    def get_server_base_url(self):
        # TODO(dmu) HIGH: It is strange that we use slightly different way to get `server_base_url`
        #                 for bootstrap server and packy agent
        return self.get_packy_agent_configuration().get('server_base_url')

    def get_server_hostname(self):
        return get_hostname_from_url(self.get_server_base_url())

    def get_agent_id(self):
        return self.get_packy_agent_configuration().get('agent_id')

    def get_client_id(self):
        return self.get_packy_agent_configuration().get('client_id')

    def get_client_secret(self):
        return self.get_packy_agent_configuration().get('client_secret')

    def is_activated(self):
        return (self.get_agent_id() is not None and self.get_client_id() is not None and
                self.get_client_secret() is not None)

    def deactivate(self):
        self.load_local_configuration()
        config = self.local_configuration.setdefault('packy', {}).setdefault('agent', {})

        config['agent_id'] = None
        config['client_id'] = None
        config['client_secret'] = None
        self.save_local_configuration()

    def get_tasks_prefix(self):
        return self.get_packy_agent_configuration().get('tasks_prefix')

    def get_module_configuration(self, module_public_identifier):
        return (self.get_packy_agent_configuration().get('modules') or {}).get(
            module_public_identifier)

    def get_ping_module_configuration(self):
        return self.get_module_configuration('ping')

    def get_trace_module_configuration(self):
        return self.get_module_configuration('trace')

    def get_speedtest_module_configuration(self):
        return self.get_module_configuration('speedtest')

    def get_http_module_configuration(self):
        return self.get_module_configuration('http')

    def is_debug_mode(self):
        return (self.get_packy_agent_configuration().get('is_debug_mode') or
                super(PackyAgentConfiguration, self).is_debug_mode())

    def set_stopped(self, value):
        # TODO(dmu) MEDIUM: This is not 100% consistent solution. If configuration file get updated
        #                   from server `is_stopped` flag will be removed. To address this we need
        #                   either: 1) save `is_stopped` to server 2) have another storage (config
        #                   file or local database) to store data that cannot be overwritten from
        #                   server
        self.load_local_configuration()
        self.local_configuration.setdefault('packy', {}).setdefault(
            'agent', {})['is_stopped'] = value
        self.save_local_configuration()

    def is_stopped(self):
        return self.get_packy_agent_configuration().get('is_stopped')

    def is_heartbeat_enabled(self):
        return self.get_packy_agent_configuration().get('is_heartbeat_enabled')

    def get_submit_retries(self):
        return self.get_packy_agent_configuration().get('submit_retries')

    def get_data_usage_file(self):
        return self.get_packy_agent_configuration().get('data_usage_file')

    def get_packy_agent_sentry_dsn(self):
        return self.get_packy_agent_configuration().get('packy_agent_sentry_dsn')


configuration = PackyAgentConfiguration()



import argparse

from packy_agent.watchdog.service import WatchdogService
from packy_agent.utils.sentry import get_raven_client


def entry():
    raven_client = get_raven_client('watchdog')
    from packy_agent.watchdog import run as the_module

    parser = argparse.ArgumentParser(
        # Cannot use __name__, because when module is run as script it is equal to __main__
        prog='python -m {}'.format(the_module.__name__),
        formatter_class=argparse.ArgumentDefaultsHelpFormatter)

    parser.parse_args()

    service = WatchdogService(raven_client=raven_client)
    service.run()


if __name__ == '__main__':
    entry()



import argparse
import sys

from packy_agent.managers.network import network_manager


def set_dhcp(interface):
    network_manager.set_dhcp(interface)


def entry():
    from packy_agent.cli import set_dhcp as the_module
    parser = argparse.ArgumentParser(
        # Cannot use __name__, because when module is run as script it is equal to __main__
        prog='python -m {}'.format(the_module.__name__),
        formatter_class=argparse.ArgumentDefaultsHelpFormatter)

    parser.add_argument('-i', '--interface')
    args = parser.parse_args()

    return set_dhcp(args.interface)


if __name__ == '__main__':
    sys.exit(entry())



import logging

from celery import shared_task
from retry.api import retry_call

from packy_agent.clients.packy_server import packy_server_client_factory
from packy_agent.configuration.agent.base import configuration
from packy_agent.modules.speedtest.base import run_speedtest
from packy_agent.utils.misc import get_iso_format_utcnow
from packy_agent.tasks.modules.base import run_exclusive

MBIT = 1000000.0

logger = logging.getLogger(__name__)


def speedtest(host):
    start_time = get_iso_format_utcnow()
    results = run_speedtest(host)
    # TODO(dmu) HIGH: Producing of result structure should be moved to PackyServerClient level
    return {
        'target': host,
        'upload_speed': round(results.upload / MBIT, 2),
        'download_speed': round(results.download / MBIT, 2),
        'time': start_time,
        'ping': round(results.ping, 3)
    }


def speedtest_all_impl():
    # TODO(dmu) HIGH: Reloading configuration file at every task run is a bad idea. Refactor
    hosts = configuration.get_speedtest_module_configuration().get('hosts')
    if not hosts:
        logger.warning('Speedtest hosts are not configured')
        # TODO(dmu) HIGH: Should I mark task as failed?
        return

    client = packy_server_client_factory.get_client_by_configuration(configuration)

    # Celery has limitations for creating subprocesses via multiprocessing
    # https://github.com/celery/celery/issues/1709 therefore we run speedtest measurement
    # sequentially
    for host in hosts:
        result = speedtest(host)
        retry_call(client.submit_speedtest_measurement, (result,),
                   tries=configuration.get_submit_retries())


@shared_task
def speedtest_all():
    run_exclusive(speedtest_all_impl)



from __future__ import absolute_import

import logging
import os
import socket
import time
import select
from collections import defaultdict
from contextlib import closing
from collections import namedtuple

from packy_agent.modules.base.socket import get_seq_no, send_icmp_echo_request as send_icmp_probe
from packy_agent.modules.traceroute.constants import ICMP_METHOD, UDP_METHOD
from packy_agent.modules.traceroute.methods.icmp import (
    PACKET_HEADERS_SIZE_BYTES as ICMP_METHOD_PACKET_HEADERS_SIZE_BYTES)
from packy_agent.modules.traceroute.methods.udp import send_udp_probe
from packy_agent.modules.traceroute.types.parallel.udp import process_udp_probe
from packy_agent.modules.traceroute.types.parallel.icmp import process_icmp_probe

logger = logging.getLogger(__name__)
SentProbe = namedtuple('SentProbe', ('ttl', 'probe_number', 'sent_time', 'payload'))


def get_ttl_and_probe_number(max_hops, probe_count):
    for ttl in xrange(1, max_hops + 1):
        for probe_number in xrange(probe_count):
            yield (ttl, probe_number)


def receive_probe(method, sent_probes, icmp_socket, id_=None, timeout=0.001):
    ready_sockets = select.select((icmp_socket,), (), (), timeout)
    # because by this time the entire packet is already in OS buffer
    finish_time = time.time()

    if ready_sockets[0]:
        if method == UDP_METHOD:
            return_value = process_udp_probe(sent_probes, icmp_socket, finish_time)
        elif method == ICMP_METHOD:
            return_value = process_icmp_probe(sent_probes, id_, icmp_socket, finish_time)
        else:
            raise ValueError('Unknown traceroute method: {}'.format(method))
    else:  # timeout
        return_value = None

    return return_value


def traceroute_parallel_impl(method, send_socket, icmp_socket,
                             kwargs, destination_ip_address, probe_count,
                             max_hops=100, timeout=2, max_parallelism=10):
    if method == UDP_METHOD:
        # TODO(dmu) LOW: Improve to reuse ports once they are not needed any more
        port = 33434
        receive_probe_kwargs = {}
    elif method == ICMP_METHOD:
        id_ = os.getpid() & 0xFFFF
        receive_probe_kwargs = {'id_': id_}
    else:
        raise ValueError('Unknown traceroute method: {}'.format(method))

    destination_hop_number = max_hops
    sent_probes = {}
    results = defaultdict(dict)
    is_done = False
    should_send = True
    ttl_and_probe_number_generator = get_ttl_and_probe_number(max_hops, probe_count)
    while not is_done:
        if should_send and len(sent_probes) < max_parallelism:
            try:
                ttl, probe_number = next(ttl_and_probe_number_generator)
                should_send = ttl <= destination_hop_number
            except StopIteration:
                should_send = False

            if should_send:
                if method == UDP_METHOD:
                    sent_time, sent_payload = send_udp_probe(
                        send_socket, destination_ip_address=kwargs['destination_ip_address'],
                        port=port, ttl=ttl, packet_size_bytes=kwargs['packet_size_bytes'])
                    sent_probes[port] = SentProbe(ttl, probe_number, sent_time, sent_payload)
                    port += 1
                elif method == ICMP_METHOD:
                    icmp_socket.setsockopt(socket.SOL_IP, socket.IP_TTL, ttl)
                    seq_no = get_seq_no('icmp')
                    sent_time, sent_payload = send_icmp_probe(
                        send_socket, destination_ip_address, id_, seq_no,
                        kwargs['packet_size_bytes'] - ICMP_METHOD_PACKET_HEADERS_SIZE_BYTES)
                    sent_probes[seq_no] = SentProbe(ttl, probe_number, sent_time, sent_payload)
                else:
                    raise ValueError('Unknown traceroute method: {}'.format(method))

        while True:
            received_probe = receive_probe(
                method, sent_probes, icmp_socket, **receive_probe_kwargs)

            if not received_probe:
                break

            results[received_probe.hop_number][received_probe.probe_number] = received_probe

            if (received_probe.is_destination_reached or
                    received_probe.hop_ip_address == destination_ip_address):
                destination_hop_number = min(received_probe.hop_number, destination_hop_number)

            for ttl_local in xrange(destination_hop_number, 0, -1):
                if len(results[ttl_local]) < probe_count:
                    break
            else:
                is_done = True
                break

        for sent_probe_key, sent_probe in sent_probes.items():
            if sent_probe.sent_time + timeout <= time.time():
                if method == UDP_METHOD:
                    logger.debug('Probe on port {} timed out'.format(sent_probe_key))
                elif method == ICMP_METHOD:
                    logger.debug('Probe with seq {} timed out'.format(sent_probe_key))
                else:
                    raise ValueError('Unknown traceroute method: {}'.format(method))

                del sent_probes[sent_probe_key]

        if not should_send and not sent_probes:
            is_done = True

    return [[results[ttl].get(probe_number) for probe_number in xrange(probe_count)]
            for ttl in xrange(1, destination_hop_number + 1)]


def traceroute_parallel(method, kwargs, destination_ip_address, probe_count,
                        max_hops=100, timeout=2, max_parallelism=10):

    udp_protocol = socket.getprotobyname('udp')
    icmp_protocol = socket.getprotobyname('icmp')

    try:
        with closing(socket.socket(socket.AF_INET, socket.SOCK_RAW, icmp_protocol)) as icmp_socket:
            if method == UDP_METHOD:
                with closing(socket.socket(
                        socket.AF_INET, socket.SOCK_DGRAM, udp_protocol)) as udp_socket:
                    return traceroute_parallel_impl(
                        method, send_socket=udp_socket, icmp_socket=icmp_socket,
                        kwargs=kwargs, destination_ip_address=destination_ip_address,
                        probe_count=probe_count, max_hops=max_hops, timeout=timeout,
                        max_parallelism=max_parallelism)
            elif method == ICMP_METHOD:
                return traceroute_parallel_impl(
                    method, send_socket=icmp_socket, icmp_socket=icmp_socket,
                    kwargs=kwargs, destination_ip_address=destination_ip_address,
                    probe_count=probe_count, max_hops=max_hops, timeout=timeout,
                    max_parallelism=max_parallelism)
            else:
                raise ValueError('Unknown traceroute method: {}'.format(method))

    except socket.error as (errno, msg):
        if errno == 1:  # Operation not permitted
            # TODO(dmu) HIGH: Fix to receive ICMP packet without being root
            msg += ' - process should be running as root.'
            raise socket.error(msg)

        raise



import argparse

from packy_agent.control_server.app import get_app


def entry():
    from packy_agent.control_server import run as the_module

    parser = argparse.ArgumentParser(
        # Cannot use __name__, because when module is run as script it is equal to __main__
        prog='python -m {}'.format(the_module.__name__),
        formatter_class=argparse.ArgumentDefaultsHelpFormatter)

    parser.add_argument('--host', default='127.0.0.1')
    parser.add_argument('--port', type=int, default=8001)
    parser.add_argument('--debug', action='store_true')

    args = parser.parse_args()

    app = get_app()
    app.run(host=args.host, port=args.port, debug=args.debug)


if __name__ == '__main__':
    entry()



import argparse
import sys

from packy_agent.managers.install_and_upgrade import install_and_upgrade_manager, CONFIGURE_TARGETS
from packy_agent.utils.logging import KNOWN_LOG_LEVELS, configure_basic_logging


# TODO(dmu) HIGH: Stopping keeping this code once all agent upgrade to v0.3.4 or higher
def entry():
    from packy_agent.cli import configure as the_module
    parser = argparse.ArgumentParser(
        # Cannot use __name__, because when module is run as script it is equal to __main__
        prog='python -m {}'.format(the_module.__name__),
        formatter_class=argparse.ArgumentDefaultsHelpFormatter)

    parser.add_argument('target', choices=CONFIGURE_TARGETS)
    parser.add_argument('base_url')
    parser.add_argument('--log-level', default='DEBUG', choices=KNOWN_LOG_LEVELS)
    args = parser.parse_args()

    configure_basic_logging(level=args.log_level)

    install_and_upgrade_manager.install_and_restart()


if __name__ == '__main__':
    sys.exit(entry())



import logging
import os
import os.path

from packy_agent.managers.network import INTERFACES_CONFIG_PATH, BACKUP_FILENAME_TEMPLATE
from packy_agent.configuration.agent.base import configuration as agent_configuration
from packy_agent.configuration.control_server.base import configuration as control_server_configuration
from packy_agent.utils.misc import is_inside_docker_container
from packy_agent.clients.packy_server import packy_server_client_factory


logger = logging.getLogger(__name__)


class ResetManager(object):

    def reset_network_configuration(self):
        if control_server_configuration.is_network_configuration_enabled():
            backup_file = BACKUP_FILENAME_TEMPLATE.format(INTERFACES_CONFIG_PATH)
            try:
                os.rename(backup_file, INTERFACES_CONFIG_PATH)
            except Exception as ex:
                logger.warning('Network configuration was not restored: %s', ex)
        else:
            logger.info('Network configuration change was disabled (for developer protection)')

    def deactivate(self):
        client = packy_server_client_factory.get_client_by_configuration(agent_configuration)
        client.deactivate_agent()

        # We are doing deactivate() as a defensive programming measure to avoid accident
        # overwriting deleted file with configuration from memory
        agent_configuration.deactivate()

        # We remove file to be closer to fresh install state
        file_path = agent_configuration.get_local_configuration_file_path()
        logger.info('Removing configuration file: %s', file_path)
        if os.path.isfile(file_path):
            try:
                os.remove(file_path)
            except Exception:
                logger.exception('Error while removing configuration file: %s', file_path)

    def reset_all(self):
        if not is_inside_docker_container():
            self.reset_network_configuration()

        self.deactivate()


reset_manager = ResetManager()



import logging

from celery import shared_task
from retry.api import retry_call

from packy_agent.clients.packy_server import packy_server_client_factory
from packy_agent.configuration.agent.base import configuration
from packy_agent.modules.traceroute.base import traceroute
from packy_agent.tasks.modules.base import run_exclusive
from packy_agent.modules.traceroute.constants import UDP_METHOD
from packy_agent.utils.misc import get_iso_format_utcnow

logger = logging.getLogger(__name__)


def convert_to_hops(traceroute_result, probe_count):
    hops = []
    for x, item in enumerate(traceroute_result):
        hop = {
            'host': item.get('host'),
            'number': x + 1,
            'reply_number': item.get('reply_hop_number'),
            'loss': round(100.0 * item['loss'] / probe_count, 1),
            'sent': probe_count,
        }
        for attr in ('last', 'average', 'best', 'worst', 'stdev'):
            value = item.get(attr)
            if value is not None:
                hop[attr] = round(value * 1000.0, 1)

        hops.append(hop)

    return hops


def trace(host):
    module_config = configuration.get_trace_module_configuration()

    start_time = get_iso_format_utcnow()
    probe_count = module_config['number_of_pings']
    packet_size_bytes = module_config['packet_size']
    if not (28 <= packet_size_bytes <= 1500):
        packet_size_bytes = min(max(28, packet_size_bytes), 1500)
        logger.warning('`packet_size` is out of range from 28 to 1500 (inclusive), '
                       'value is adjusted to {}'.format(packet_size_bytes))

    result = traceroute(host, probe_count=probe_count,
                        packet_size_bytes=packet_size_bytes,
                        max_hops=module_config['ttl'],
                        method=module_config.get('method', UDP_METHOD),
                        max_parallelism=module_config.get('parallelism', 16))
    if result is None:
        return

    # TODO(dmu) HIGH: Producing of result structure should be moved to PackyServerClient level
    return {
        'target': host,
        'packet_size': module_config['packet_size'],
        'pings': module_config['number_of_pings'],
        'time': start_time,
        'hops': convert_to_hops(result, probe_count),
    }


def trace_all_impl():
    hosts = configuration.get_trace_module_configuration().get('hosts')
    if not hosts:
        logger.warning('Trace hosts are not configured')
        # TODO(dmu) HIGH: Should I mark task as failed?
        return

    client = packy_server_client_factory.get_client_by_configuration(configuration)
    # See comment for speedtest
    for host in hosts:
        try:
            result = trace(host)
            if result is None:
                continue
        except Exception:
            logger.warning('Exception during trace: %s', host)
            continue

        retry_call(client.submit_trace_measurement, (result,),
                   tries=configuration.get_submit_retries())


@shared_task
def trace_all():
    run_exclusive(trace_all_impl)



from __future__ import absolute_import

import logging
import math
import socket
from functools import partial
from collections import namedtuple

from packy_agent.modules.traceroute.constants import (
    ICMP_METHOD, UDP_METHOD)
from packy_agent.modules.base.constants import IP_PACKET_HEADER_SIZE_BYTES, \
    ICMP_PACKET_HEADER_SIZE_BYTES, UDP_PACKET_HEADER_SIZE_BYTES
from packy_agent.modules.traceroute.methods.icmp import trace_hop_with_icmp
from packy_agent.modules.traceroute.methods.udp import trace_hop_with_udp
from packy_agent.modules.traceroute.types.parallel.base import traceroute_parallel
from packy_agent.modules.traceroute.types.sequential import traceroute_sequential

ProbeResult = namedtuple('ProbeResult', ('hop_number', 'reply_hop_number', 'probe_number',
                                         'hop_ip_address', 'rtt_seconds', 'is_destination_reached'))

MIN_ICMP_METHOD_PACKET_SIZE = IP_PACKET_HEADER_SIZE_BYTES + ICMP_PACKET_HEADER_SIZE_BYTES
MIN_UDP_METHOD_PACKET_SIZE = IP_PACKET_HEADER_SIZE_BYTES + UDP_PACKET_HEADER_SIZE_BYTES
RANGE_MIN = {
    ICMP_METHOD: MIN_ICMP_METHOD_PACKET_SIZE,
    UDP_METHOD: MIN_UDP_METHOD_PACKET_SIZE,
}
MAX_PACKET_SIZE = 1500

logger = logging.getLogger(__name__)


def aggregate_results(trace_results):
    agg_results = []
    for probe_results in trace_results:
        aggregated = {'loss': probe_results.count(None)}
        filtered_probe_results = filter(None, probe_results)
        if filtered_probe_results:
            # TODO(dmu) HIGH: Host may be different for the different probes of the same TTL
            representative_probe = filtered_probe_results[0]
            aggregated['host'] = representative_probe.hop_ip_address
            aggregated['reply_hop_number'] = representative_probe.reply_hop_number
            rtts = tuple(x.rtt_seconds for x in filtered_probe_results)
            aggregated['last'] = rtts[-1]
            aggregated['best'] = min(rtts)
            aggregated['worst'] = max(rtts)
            average = sum(rtts) / len(rtts)
            aggregated['average'] = average

            variance = sum((x - average) ** 2 for x in rtts) / len(rtts)
            aggregated['stdev'] = math.sqrt(variance)

        agg_results.append(aggregated)

    return agg_results


def traceroute(host, timeout=2, probe_count=1, packet_size_bytes=60, max_hops=100,
               method=UDP_METHOD, max_parallelism=10):

    if method not in (UDP_METHOD, ICMP_METHOD):
        raise ValueError('Unknown traceroute method: {}'.format(method))

    if max_parallelism < 1:
        raise ValueError('max_parallelism must be greater or equal to 1')

    range_min = RANGE_MIN[method]
    if not (range_min <= packet_size_bytes <= MAX_PACKET_SIZE):
        raise ValueError(
            'packet_size_bytes must be in range from {} to 1500 (inclusive)'.format(range_min))

    try:
        destination_ip_address = socket.gethostbyname(host)
    except socket.gaierror:
        logger.warning('Could not resolve {} to IP address'.format(host))
        return

    kwargs = {
        'destination_ip_address': destination_ip_address,
        'timeout': timeout,
        'packet_size_bytes': packet_size_bytes,
    }

    if max_parallelism == 1:
        if method == ICMP_METHOD:
            trace_function = trace_hop_with_icmp
        elif method == UDP_METHOD:
            port = [33434]

            def get_port():
                port[0] += 1
                return int(port[0])

            kwargs['port'] = get_port

            trace_function = trace_hop_with_udp
        else:
            assert False, 'Should never get here'
            raise ValueError('Unknown traceroute method: {}'.format(method))

        trace_function_partial = partial(trace_function, **kwargs)
        results = traceroute_sequential(trace_function_partial, destination_ip_address,
                                        probe_count=probe_count, max_hops=max_hops)
    else:
        results = traceroute_parallel(method, kwargs, destination_ip_address,
                                      probe_count=probe_count,
                                      max_hops=max_hops, timeout=timeout,
                                      max_parallelism=max_parallelism)

    return aggregate_results(results)



DEFAULT_SERVER_BASE_URL = 'https://p04.packy.io/'
PACKY_SERVER_BASE_URL_ENVVAR_NAME = 'PACKY_SERVER_BASE_URL'
AGENT_CONFIG_ENVVAR_NAME = 'PACKY_AGENT_CONFIG'
DEFAULT_AGENT_CONFIG_PATH = '/etc/packy-agent.yaml'
CONTROL_SERVER_CONFIG_ENVVAR_NAME = 'PACKY_CONTROL_SERVER_CONFIG'
DEFAULT_CONTROL_SERVER_CONFIG_PATH = '/etc/packy-control-server.yaml'



import logging

import flask
from flask.views import MethodView

from packy_agent.managers.control import control_manager
from packy_agent.control_server.forms.action import ActionForm
from packy_agent.control_server.views.base import smart_redirect
from packy_agent.utils.auth import activation_and_authentication_required


DELAY_SECONDS = 5

logger = logging.getLogger(__name__)


def do_action(action):
    if action == 'reboot':
        flask.flash('Reboot will start in {} seconds...'.format(DELAY_SECONDS))
        control_manager.reboot(delay_seconds=DELAY_SECONDS)
    elif action == 'start_agent':
        control_manager.start_packy_agent()
    elif action == 'stop_agent':
        control_manager.stop_packy_agent()
    elif action == 'restart_agent':
        control_manager.restart_packy_agent()
    elif action == 'restart':
        # Restart is delayed because we need to send HTTP response before Control Server goes down
        flask.flash('Restart will start in {} seconds...'.format(DELAY_SECONDS))
        control_manager.restart_all(delay_seconds=DELAY_SECONDS)
    elif action == 'restart_control_server':
        # Restart is delayed because we need to send HTTP response before Control Server goes down
        flask.flash('Restart will start in {} seconds...'.format(DELAY_SECONDS))
        control_manager.restart_control_server(delay_seconds=DELAY_SECONDS)
    else:
        raise ValueError('Unknown action: {}'.format(action))


class ControlView(MethodView):

    @activation_and_authentication_required
    def get(self):
        form = ActionForm()

        agent_status = control_manager.get_packy_agent_status()
        agent_status_line = agent_status[0] or 'UNKNOWN'
        if agent_status[1]:
            agent_status_line += ' (uptime {})'.format(agent_status[1])

        context = {
            'form': form,
            'agent_status': agent_status_line,
            'active_menu_item': 'control',
        }

        return flask.render_template('control.html', **context)

    @activation_and_authentication_required
    def post(self):
        form = ActionForm()
        if form.validate():
            action = flask.request.form['action']
            do_action(action)
            return smart_redirect('success', 'control')

        return flask.render_template('control.html', form=form)



from __future__ import absolute_import


import logging
import os
import socket
import select
import time

from ryu.lib.packet.icmp import ICMP_ECHO_REPLY
from ryu.lib.packet.ipv4 import ipv4
from ryu.lib.packet.packet import Packet

from contextlib import closing
from packy_agent.modules.base.socket import (get_seq_no, TIME_SIZE_BYTES,
                                             send_icmp_echo_request, decode_timestamp)


logger = logging.getLogger(__name__)


def receive_icmp_reply(raw_socket, id_, seq_no, payload, timeout, sent_time):
    payload_len = len(payload)

    time_left = timeout
    while time_left > 0:
        start_time = time.time()
        ready_sockets = select.select((raw_socket,), (), (), time_left)
        finish_time = time.time()  # because by this time the entire packet is already in OS buffer
        time_left -= (finish_time - start_time)

        if not ready_sockets[0]:  # timeout
            break

        ip_packet_data, (ip_address, _) = raw_socket.recvfrom(1024 + payload_len)

        packet = Packet(ip_packet_data, parse_cls=ipv4)

        icmp_packet = packet.protocols[1]
        logger.debug('Got ICMP packet type:{} code:{} from {}'.format(
            icmp_packet.type, icmp_packet.code, ip_address))
        if icmp_packet.type != ICMP_ECHO_REPLY:
            logger.debug('Discarded ICMP packet type:{} code:{} from {}'.format(
                icmp_packet.type, icmp_packet.code, ip_address))
            continue

        packet_seq = icmp_packet.data.seq
        packet_id = icmp_packet.data.id
        packet_payload = icmp_packet.data.data

        logger.debug(
            'ICMP packet from {} type:{} code:{} has id:{}, seq:{}, payload:{}'.format(
            ip_address, icmp_packet.type, icmp_packet.code, packet_id, packet_seq,
            packet_payload.encode('hex') if packet_payload else packet_payload))

        if packet_seq != seq_no or packet_id != id_:
            continue

        if packet_payload:
            packet_payload_len = len(packet_payload)
            if packet_payload_len != payload_len:
                logger.warning('Received payload length ({} bytes) differ from expected (sent) '
                               'length ({} bytes), but we are forgiving'.format(
                                packet_payload_len, payload_len))
                min_len = min(packet_payload_len, payload_len)
                if packet_payload[:min_len] != payload[:min_len]:
                    logger.debug('Discarded ICMP packet type:{} code:{} from {}: '
                                 'payload prefix differs'.format(icmp_packet.type,
                                                                 icmp_packet.code, ip_address))
                    continue
            elif packet_payload != payload:
                logger.debug('Discarded ICMP packet type:{} code:{} from {}: '
                             'payload differs'.format(icmp_packet.type, icmp_packet.code,
                                                      ip_address))
                continue

        if packet_payload and len(packet_payload) >= TIME_SIZE_BYTES:
            sent_time = decode_timestamp(packet_payload[:TIME_SIZE_BYTES])

        return finish_time - sent_time


def ping_once(destination_ip_address, timeout, packet_size_bytes=56):
    icmp_protocol = socket.getprotobyname('icmp')
    try:
        with closing(socket.socket(socket.AF_INET, socket.SOCK_RAW, icmp_protocol)) as raw_socket:
            id_ = os.getpid() & 0xFFFF
            seq_no = get_seq_no('icmp')
            sent_time, payload = send_icmp_echo_request(
                raw_socket, destination_ip_address, id_, seq_no, packet_size_bytes)
            return receive_icmp_reply(raw_socket, id_, seq_no, payload, timeout,
                                      sent_time=sent_time)

    except socket.error as (errno, msg):
        if errno == 1:  # Operation not permitted
            msg += ' - Note that ICMP messages can only be sent from processes running as root.'
            raise socket.error(msg)

        raise



import logging
import os
import os.path
from urllib import urlencode

from filelock import FileLock, Timeout
from jinja2 import Template
from pkg_resources import parse_version

from packy_agent.clients.packy_server import PackyServerClient
from packy_agent.clients.packy_server import packy_server_client_factory
from packy_agent.configuration.agent.base import configuration as packy_agent_configuration
from packy_agent.configuration.control_server.base import configuration
from packy_agent.exceptions import AuthenticationError, ValidationError
from packy_agent.managers.control import control_manager
from packy_agent.utils.auth import is_activated
from packy_agent.utils.misc import run_shell_command_async, is_inside_docker_container, \
    UPGRADE_IN_PROGRESS_FILE_LOCK
from packy_agent.utils.network import get_machine_ip_address, get_hostname_from_url
from packy_agent.utils.output import write_to_console_or_file
from packy_agent.utils.pkg_resources import get_package_file_content

logger = logging.getLogger(__name__)

UBUNTU_CONFIGURE_TARGET = 'ubuntu'
DOCKER_CONFIGURE_TARGET = 'docker'
CONFIGURE_TARGETS = (UBUNTU_CONFIGURE_TARGET, DOCKER_CONFIGURE_TARGET)
PACKY_GET_INSTALL_SCRIPT_PATH = '/tmp/packy-get-install-script.sh'


def render_template_from_package_file_content(template_module_name, template_filename,
                                              context=None):
    template_text = get_package_file_content(template_module_name, template_filename)
    template = Template(template_text)
    return template.render(**(context or {}))


class InstallAndUpgradeManager(object):

    def activate(self, email, password, agent_id=None, async_restart=True):
        if is_activated():
            raise ValidationError('Agent has already been activated')

        client = PackyServerClient(configuration.get_server_base_url(), agent_id=agent_id)
        if agent_id is None:
            response = client.create_agent(email, password)
        else:
            response = client.activate_agent(email, password)
        status_code = response.status_code
        if status_code == 401:
            raise AuthenticationError('Invalid credentials for agent activation')
        elif status_code == 400:
            payload = response.json()
            errors = payload.get('errors') or {}
            non_field_errors = errors.pop('non_field_errors', None) or []
            messages = ([payload.get('message')] + non_field_errors +
                        ['{}: {}.'.format(k, v) for k, v in errors.iteritems()])

            raise ValidationError(' '.join(filter(None, messages)))

        response.raise_for_status()

        agent_configuration = response.json()['configuration']
        packy_agent_configuration.save_local_configuration(agent_configuration)

        # TODO(dmu) LOW: Instead of always restart do according to current running status
        control_manager.restart_packy_agent(delay_seconds=1 if async_restart else None)

    def install_and_restart(self, version=None, delay_seconds=None):
        logger.info('Upgrading...')
        if not configuration.is_upgrade_enabled():
            logger.info('Upgrade was disabled (for developer protection)')
            return

        file_lock = FileLock(UPGRADE_IN_PROGRESS_FILE_LOCK)
        try:
            with file_lock.acquire(timeout=1):
                client = packy_server_client_factory.get_client_by_configuration(
                    packy_agent_configuration)
                version_max = client.get_version_max()
                if version is None:
                    version = version_max
                else:
                    if version_max and parse_version(version) > parse_version(version_max):
                        raise ValueError('Could not install/upgrade to version {} '
                                         '(it is higher than {})'.format(version, version_max))

                # TODO(dmu) HIGH: Improve OS detection
                operating_system = 'alpine' if is_inside_docker_container() else 'ubuntu'
                download_args = {'os': operating_system}
                if version:
                    download_args['version'] = version

                write_to_console_or_file(
                    PACKY_GET_INSTALL_SCRIPT_PATH,
                    render_template_from_package_file_content(
                        'packy_agent', 'scripts/misc/packy-get-install-script.sh.j2',
                        {'with_shebang': True,
                         'download_args': '?{}'.format(urlencode(download_args))}))

                # Restart is done in the install script
                extra_envvars = {'PACKY_SERVER_BASE_URL': "'{}'".format(
                    configuration.get_server_base_url().rstrip('/'))}

                packy_agent_pypi_repository = configuration.get_packy_agent_pypi_repository()
                if packy_agent_pypi_repository:
                    extra_envvars['PIP_EXTRA_ARGS'] = "'--extra-index-url {}'".format(
                        packy_agent_pypi_repository)

                if os.getenv('OLD_PATH'):
                    extra_envvars['PATH'] = '$OLD_PATH'

                command = '{} /bin/bash {}; rm -f {}'.format(
                    ' '.join('export {}={};'.format(k, v) for k, v in extra_envvars.iteritems()),
                    PACKY_GET_INSTALL_SCRIPT_PATH, UPGRADE_IN_PROGRESS_FILE_LOCK)

                run_shell_command_async(command, delay_seconds=delay_seconds)
        except Timeout:
            logger.warning('Concurrent upgrade detected')

    def upgrade_and_restart(self, delay_seconds=None):
        self.install_and_restart(delay_seconds=delay_seconds)

    # TODO(dmu) MEDIUM: Refactor to use Ansible for most of these stuff
    def get_control_server_link(self, packy_server_base_url):
        # TODO(dmu) LOW: Unhardcode schema: `http://`
        link = 'http://' + get_machine_ip_address(get_hostname_from_url(packy_server_base_url))
        port = configuration.get_control_server_port()
        if port != 80:
            link += ':{}'.format(port)

        return link


install_and_upgrade_manager = InstallAndUpgradeManager()



import time
import logging
import signal

from uptime import uptime
import psutil

from packy_agent.configuration.agent.base import configuration as agent_configuration
from packy_agent.configuration.control_server.base import configuration as cs_configuration
from packy_agent.clients.packy_server import packy_server_client_factory
from packy_agent.utils.misc import (to_epoch, iso_to_datetime, is_inside_docker_container,
                                    dump_version)
from packy_agent.managers.control import control_manager
from packy_agent.managers.install_and_upgrade import install_and_upgrade_manager
from packy_agent.utils.sentry import get_raven_client


logger = logging.getLogger(__name__)


FAILOVER_CHECK_PERIOD_SECONDS = 30
ROOT_PID = 1


def custom_uptime():
    if is_inside_docker_container():
        process = psutil.Process(ROOT_PID)
        return time.time() - process.create_time()
    else:
        return uptime()


class WatchdogService(object):

    def __init__(self, raven_client=None):
        self.raven_client = raven_client

        self._graceful_stop = False

        self.known_activated_time = None
        self.last_restart_time = None
        self.last_known_online_time = None

    def get_raven_client(self):
        if not self.raven_client:
            self.raven_client = get_raven_client('watchdog')

        return self.raven_client

    def capture_exception(self):
        raven_client = self.get_raven_client()
        if raven_client:
            raven_client.captureException()

    def capture_message(self, message, level=logging.INFO):
        raven_client = self.get_raven_client()
        if raven_client:
            raven_client.captureMessage(message, data={'level': level})

    def capture_warning(self, message):
        self.capture_message(message, level=logging.WARNING)

    def graceful_stop(self):
        self._graceful_stop = True

    def get_last_online_time(self):
        last_online_dt = None
        try:
            client = packy_server_client_factory.get_client_by_configuration(agent_configuration)
            response = client.get_agent()
            body_json = response.json()
            last_online_dt = body_json.get('last_online_dt')
        except Exception:
            logger.exception('Error while getting online time')

        if last_online_dt:
            last_online_time = to_epoch(iso_to_datetime(last_online_dt))
            self.last_known_online_time = last_online_time
            return last_online_time
        else:
            if self.last_known_online_time is None:
                return self.known_activated_time
            else:
                return self.last_known_online_time

    def get_offline_period(self):
        return time.time() - self.get_last_online_time()

    def notify_server(self):
        base_url = cs_configuration.get_server_base_url()

        client = packy_server_client_factory.get_client(base_url)
        control_server_url = install_and_upgrade_manager.get_control_server_link(base_url)
        client.notify_server(control_server_url)

    def loop_iteration(self):
        if not agent_configuration.is_activated():
            logger.info('Agent is not activated')
            self.notify_server()
            return

        if agent_configuration.is_stopped():
            logger.info('Agent is stopped on purpose')
            return

        if self.known_activated_time is None:
            self.known_activated_time = time.time()

        offline_period = self.get_offline_period()
        watchdog_config = agent_configuration.get_watchdog_configuration()
        if watchdog_config['max_offline_period_before_reboot_seconds'] < offline_period:
            if custom_uptime() >= watchdog_config['reboot_period_seconds']:
                self.capture_warning('Watchdog requested reboot')
                control_manager.reboot()
                return

        if watchdog_config['max_offline_period_before_restart_seconds'] < offline_period:
            if (not self.last_restart_time or
                    (time.time() - self.last_restart_time) >=
                        watchdog_config['restart_period_seconds']):
                self.last_restart_time = time.time()
                self.capture_warning('Watchdog requested restart')
                control_manager.restart_packy_agent()
                return

    def run(self):
        # TODO(dmu) HIGH: Improve logging configuration
        logging.basicConfig(format='%(asctime)-15s %(levelname)s [%(name)s]: %(message)s',
                            level=logging.DEBUG)

        logger.info('Watchdog started')
        signal.signal(signal.SIGTERM, lambda signum, frame: self.graceful_stop())
        while not self._graceful_stop:
            start = time.time()
            try:
                self.loop_iteration()
            except Exception:
                logger.exception('Error during loop iteration')
                self.capture_exception()

            # Configuration may change with time, so we reread it every loop
            try:
                check_period_seconds = agent_configuration.get_watchdog_configuration(
                    )['check_period_seconds']
            except Exception:
                check_period_seconds = FAILOVER_CHECK_PERIOD_SECONDS
                logger.exception('Could not get check period seconds from configuration file, '
                                 'using failover value: %s', check_period_seconds)
                self.capture_exception()

            dump_version('/etc/packy-watchdog-version', 'PACKY_WATCHDOG_VERSION')
            wait_duration = check_period_seconds - (time.time() - start)
            if wait_duration > 0:
                logger.debug('Waiting for %s seconds for next iteration', wait_duration)
                time.sleep(wait_duration)

        logger.info('Watchdog gracefully stopped')



import argparse
import sys
import os

from packy_agent.managers.install_and_upgrade import install_and_upgrade_manager
from packy_agent.configuration.agent.base import configuration as agent_configuration
from packy_agent.configuration.control_server.base import configuration as cs_configuration
from packy_agent.utils.logging import KNOWN_LOG_LEVELS, configure_basic_logging


ACTIVATE_SCRIPT_PATH = '/tmp/packy-activate.sh'
ACTIVATE_PACKY_AGENT_SH = """#!/usr/bin/env bash
source {venv_path}/bin/activate
packy-agent-activate
"""


def entry():
    from packy_agent.cli import welcome as the_module
    parser = argparse.ArgumentParser(
        # Cannot use __name__, because when module is run as script it is equal to __main__
        prog='python -m {}'.format(the_module.__name__),
        formatter_class=argparse.ArgumentDefaultsHelpFormatter)

    parser.add_argument('--log-level', default='INFO', choices=KNOWN_LOG_LEVELS)
    args = parser.parse_args()

    configure_basic_logging(level=args.log_level)

    print '\n' + '-' * 40 + '\n'
    print 'Welcome to Packy Agent'

    if not agent_configuration.is_activated():
        with open(ACTIVATE_SCRIPT_PATH, 'w') as f:
            f.write(ACTIVATE_PACKY_AGENT_SH.format(venv_path=os.getenv('VIRTUAL_ENV')))
        os.system('chmod u+x {}'.format(ACTIVATE_SCRIPT_PATH))

        link = install_and_upgrade_manager.get_control_server_link(
            cs_configuration.get_server_base_url())
        print ('Please, activate your Packy Agent at {} or '
               'run `sudo {}`\n'.format(link, ACTIVATE_SCRIPT_PATH))


if __name__ == '__main__':
    sys.exit(entry())



import logging

import os

from filelock import FileLock, Timeout


LOCKFILE_NAME_TEMPLATE = '/tmp/packy-agent.{}.lock'

logger = logging.getLogger(__name__)


def run_exclusive(task_function, args=(), kwargs=None, lock_name=None):
    kwargs = kwargs or {}

    file_lock = FileLock(LOCKFILE_NAME_TEMPLATE.format(lock_name or task_function.__name__))

    try:
        with file_lock.acquire(timeout=1):
            return task_function(*args, **kwargs)
    except Timeout:
        logger.warning('Parallel `{}` task detected, exiting'.format(task_function.__name__))
        try:
            file_lock.acquire(timeout=300)
            file_lock.release()
        except Timeout:
            lock_file = file_lock.lock_file
            logger.warning('Self-healing: force deletion of lock file: {}'.format(lock_file))
            try:
                os.remove(lock_file)
            except OSError:
                pass



import logging

from celery import shared_task
from packy_agent.managers.control import control_manager


logger = logging.getLogger(__name__)


@shared_task()
def restart():
    control_manager.restart_packy_agent(delay_seconds=2)



import logging

from celery import shared_task
from packy_agent.managers.install_and_upgrade import install_and_upgrade_manager


logger = logging.getLogger(__name__)


def install_impl(version=None):
    install_and_upgrade_manager.install_and_restart(version=version)


# TODO(dmu) MEDIUM: Remove `update_self`
@shared_task()
def update_self():
    install_impl()


@shared_task()
def upgrade():
    install_impl()


@shared_task()
def install(version):
    install_impl(version)





def traceroute_sequential(trace_function_partial, destination_ip_address,
                          probe_count, max_hops):
    destination_hop = max_hops
    is_destination_reached = False
    results = []
    for ttl in xrange(1, destination_hop + 1):
        results_for_ttl = []
        for probe_number in xrange(probe_count):
            received_probe = trace_function_partial(ttl=ttl)
            from packy_agent.modules.traceroute.base import ProbeResult
            if received_probe:
                received_probe = ProbeResult(
                    hop_number=ttl,
                    reply_hop_number=received_probe.reply_hop_number,
                    probe_number=probe_number,
                    hop_ip_address=received_probe.hop_ip_address,
                    rtt_seconds=received_probe.rtt_seconds,
                    is_destination_reached=received_probe.is_destination_reached)

                is_destination_reached = (
                    is_destination_reached or received_probe.is_destination_reached or
                    received_probe.hop_ip_address == destination_ip_address)

            results_for_ttl.append(received_probe)

        results.append(results_for_ttl)

        if is_destination_reached:
            break

    return results



import argparse
import logging
import sys

from packy_agent.modules.traceroute import base
from packy_agent.modules.traceroute.constants import TRACEROUTE_METHODS, UDP_METHOD


def do_traceroute(host, timeout=1, probe_count=1, packet_size_bytes=60, max_hops=30,
                  method=UDP_METHOD, max_parallelism=1):
    agg_results = base.traceroute(host, timeout=timeout, probe_count=probe_count,
                                  packet_size_bytes=packet_size_bytes, max_hops=max_hops,
                                  method=method, max_parallelism=max_parallelism)
    if not agg_results:
        print 'No results'
        return

    for x, result in enumerate(agg_results):
        print x + 1, result.get('host'), result


def entry():
    logging.basicConfig(format='%(asctime)s %(levelname)s %(message)s', level=logging.WARNING)

    from packy_agent.modules.traceroute import cli as the_module

    parser = argparse.ArgumentParser(
        # Cannot use __name__, because when module is run as script it is equal to __main__
        prog='python -m {}'.format(the_module.__name__),
        formatter_class=argparse.ArgumentDefaultsHelpFormatter)

    parser.add_argument('host')
    parser.add_argument('--max-hops', help='max hops', type=int, default=30)
    parser.add_argument('--probe-count', help='probe count', type=int, default=3)
    parser.add_argument('--packet-size-bytes', help='IP packet size in bytes',
                        type=int, default=60)
    parser.add_argument('--max-parallelism', help='max parallelism', type=int, default=1)
    parser.add_argument('--timeout', help='timout', type=float, default=1)
    parser.add_argument('-M', '--method', help='method', choices=TRACEROUTE_METHODS,
                        default=UDP_METHOD)

    args = parser.parse_args()

    return do_traceroute(args.host, timeout=args.timeout, probe_count=args.probe_count,
                         packet_size_bytes=args.packet_size_bytes,
                         max_hops=args.max_hops, method=args.method,
                         max_parallelism=args.max_parallelism)


if __name__ == '__main__':
    sys.exit(entry())



from __future__ import absolute_import

import logging
from operator import itemgetter

from logging import _levelNames

KNOWN_LOG_LEVELS = tuple(
    v for k, v in sorted(_levelNames.iteritems(), key=itemgetter(0)) if isinstance(v, basestring))

BASIC_LOGGING_FORMAT = '%(asctime)-15s %(levelname)s [%(name)s]: %(message)s'


def configure_basic_logging(format=BASIC_LOGGING_FORMAT, level=logging.DEBUG):
    logging.basicConfig(format=format, level=level)



import sys
import argparse
from getpass import getpass

from packy_agent.utils.logging import KNOWN_LOG_LEVELS, configure_basic_logging
from packy_agent.utils.auth import is_activated
from packy_agent.utils.misc import get_inactive_agents
from packy_agent.managers.install_and_upgrade import install_and_upgrade_manager
from packy_agent.exceptions import AuthenticationError, ValidationError


def entry():
    from packy_agent.cli import activate as the_module
    parser = argparse.ArgumentParser(
        # Cannot use __name__, because when module is run as script it is equal to __main__
        prog='python -m {}'.format(the_module.__name__),
        formatter_class=argparse.ArgumentDefaultsHelpFormatter)

    parser.add_argument('--email')
    parser.add_argument('--password')
    parser.add_argument('--log-level', default='ERROR', choices=KNOWN_LOG_LEVELS)
    args = parser.parse_args()

    configure_basic_logging(level=args.log_level)

    if is_activated():
        print 'Packy Agent is already activated'
        return 1

    print 'Packy Agent activation'

    email = args.email or raw_input('E-mail: ')
    password = args.password or getpass('Password: ')

    try:
        agents = get_inactive_agents(email, password)
        if agents:
            print 'Activate as:'
            print '[enter] New agent'
            for agent_id, agent_name in agents.iteritems():
                print '[id={}] {}'.format(agent_id, agent_name)

            while True:
                agent_id = raw_input('Agent id: ')
                if agent_id:
                    try:
                        agent_id = int(agent_id)
                    except (ValueError, TypeError):
                        print 'Agent id must be integer'

                    install_and_upgrade_manager.activate(email, password, agent_id=agent_id,
                                                         async_restart=False)
                else:
                    install_and_upgrade_manager.activate(email, password, async_restart=False)
    except AuthenticationError:
        print 'Not authenticated (invalid credentials)'
        return 1
    except ValidationError as ex:
        print ex.message
        return 1

    print '\n' + '-' * 40 + '\n'
    print 'Packy Agent is activated successfully\n'


if __name__ == '__main__':
    sys.exit(entry())



import argparse
import logging
import sys
import time
import socket

from packy_agent.modules.ping import base


def do_ping(host, timeout=1, ping_count=None, interval_seconds=1, packet_size_bytes=56):
    try:
        destination_ip_address = socket.gethostbyname(host)
    except socket.gaierror:
        print 'Could not resolve {} to IP address'.format(host)
        return

    print 'PING {} ({})'.format(host, destination_ip_address)
    no = 1
    while ping_count is None or ping_count > 0:
        rtt_seconds = base.ping_once(host, timeout=timeout, packet_size_bytes=packet_size_bytes)
        print '{0} time={1:.2f} ms'.format(no, round(rtt_seconds * 1000, 2))

        no += 1
        if ping_count is not None:
            ping_count -= 1

        time.sleep(interval_seconds)


def entry():
    logging.basicConfig(format='%(asctime)s %(levelname)s %(message)s', level=logging.DEBUG)

    from packy_agent.modules.traceroute import cli as the_module

    parser = argparse.ArgumentParser(
        # Cannot use __name__, because when module is run as script it is equal to __main__
        prog='python -m {}'.format(the_module.__name__),
        formatter_class=argparse.ArgumentDefaultsHelpFormatter)

    parser.add_argument('host')
    parser.add_argument('--ping-count', type=int, default=None)
    parser.add_argument('--packet-size-bytes', help='ICMP payload size in bytes',
                        type=int, default=56)
    parser.add_argument('--interval-seconds', type=float, default=1)
    parser.add_argument('--timeout', type=float, default=1)

    args = parser.parse_args()

    return do_ping(args.host, timeout=args.timeout, ping_count=args.ping_count,
                   interval_seconds=args.interval_seconds,
                   packet_size_bytes=args.packet_size_bytes)


if __name__ == '__main__':
    sys.exit(entry())



import sys
import argparse

from packy_agent.utils.logging import KNOWN_LOG_LEVELS, configure_basic_logging
from packy_agent.modules.http.base import get_http_measurement


def entry():
    from packy_agent.cli import activate as the_module
    parser = argparse.ArgumentParser(
        # Cannot use __name__, because when module is run as script it is equal to __main__
        prog='python -m {}'.format(the_module.__name__),
        formatter_class=argparse.ArgumentDefaultsHelpFormatter)

    parser.add_argument('url')
    parser.add_argument('--follow-redirects', action='store_true')
    parser.add_argument('--log-level', default='ERROR', choices=KNOWN_LOG_LEVELS)
    args = parser.parse_args()

    configure_basic_logging(level=args.log_level)

    url = args.url
    print 'Getting information for {}'.format(url)
    print get_http_measurement(url, follow_redirects=args.follow_redirects).to_json()


if __name__ == '__main__':
    sys.exit(entry())



import re
import shutil
import os.path
import logging
from io import BytesIO

from jinja2 import Template
from packy_agent.utils.misc import atomic_write

from packy_agent.utils.pkg_resources import get_package_file_content
from packy_agent.utils.misc import (get_operating_system, get_interfaces, UBUNTU_16_04_XENIAL,
                                    LINUXMINT_18_1_SERENA)
from packy_agent.configuration.control_server.base import configuration
from packy_agent.configuration.agent.base import configuration as agent_configuration
from packy_agent.utils.misc import run_shell_command_async
from packy_agent.utils.network import get_machine_ip_address


INTERFACES_CONFIG_PATH = '/etc/network/interfaces'
SUPPORTED_OPERATING_SYSTEMS = (UBUNTU_16_04_XENIAL, LINUXMINT_18_1_SERENA)
STATIC_TEMPLATE_TYPE = 'static'
DHCP_TEMPLATE_TYPE = 'dhcp'
OPERATING_SYSTEM_TEMPLATES = {
    UBUNTU_16_04_XENIAL: {STATIC_TEMPLATE_TYPE: 'templates/network/ubuntu/static',
                          DHCP_TEMPLATE_TYPE: 'templates/network/ubuntu/dhcp'},
    LINUXMINT_18_1_SERENA: {STATIC_TEMPLATE_TYPE: 'templates/network/ubuntu/static',
                            DHCP_TEMPLATE_TYPE: 'templates/network/ubuntu/dhcp'},
}
STANZA_RE = re.compile(r'^\s*(?:iface|mapping|auto|allow-\S+|source|source-directory)')
IFACE_STANZA_RE = re.compile(r'^\s*iface\s+(?P<iface>\S+)\s+inet\s+(?P<method>\S+)')
AUTO_STANZA_RE = re.compile(r'^\s*(auto|allow-auto)\s+(?P<iface>\S+)')


BACKUP_FILENAME_TEMPLATE = '{}.packy.backup'

ADDR_RE = {
    'ip_address': re.compile('\s*address\s+(\S+)'),
    'subnet_mask': re.compile('\s*netmask\s+(\S+)'),
    'default_gateway': re.compile('\s*gateway\s+(\S+)'),
    'name_servers': re.compile('\s*dns-nameservers\s+(.*)'),
}
ANY_SPACE_RE = re.compile('\s+')

logger = logging.getLogger(__name__)


def get_operating_system_template_path(template_type):
    operating_system = get_operating_system()
    return (OPERATING_SYSTEM_TEMPLATES.get(operating_system) or {}).get(template_type)


def update_network(interface, delay_seconds=0):
    run_shell_command_async('ip addr flush {interface}; '
                            'systemctl restart networking.service'.format(interface=interface),
                            delay_seconds=delay_seconds)


class NetworkManager(object):

    def __init__(self):

        self.stanzas = []
        self.interfaces_config_path = configuration.get_interfaces_config_path()

    def get_actual_ip_address(self):
        return get_machine_ip_address(agent_configuration.get_server_hostname())

    def read_network_configuration(self):
        stanzas = []

        lines = []
        stanzas.append(lines)
        with open(self.interfaces_config_path) as f:
            for line in f:
                match = STANZA_RE.match(line)
                if match:
                    lines = []
                    stanzas.append(lines)

                lines.append(line)

        self.stanzas = stanzas

    def get_stanza_by_interface(self, interface):
        for lines in self.stanzas:
            if lines:
                first_line = lines[0]
                match = IFACE_STANZA_RE.match(first_line)
                if match and match.group('iface') == interface:
                    return lines

    def get_current_configuration(self, interface):
        # TODO(dmu) MEDIUM: Implement more precise DHCP detection
        # We consider DHCP if interface is not configured in /etc/network/interfaces
        self.read_network_configuration()

        stanza = self.get_stanza_by_interface(interface)
        if not stanza:
            return {'dhcp': True}

        first_line = stanza[0]
        match = IFACE_STANZA_RE.match(first_line)
        if match.group('method') == 'dhcp':
            return {'dhcp': True}

        result = {'dhcp': False}

        for line in stanza:
            for k, regex in ADDR_RE.iteritems():
                match = regex.match(line)
                if match:
                    result[k] = match.group(1)

        name_servers = result.get('name_servers')
        if name_servers is not None:
            result['name_servers'] = ','.join(ANY_SPACE_RE.split(name_servers.strip()))

        return result

    def is_operating_system_supported(self):
        return get_operating_system() in SUPPORTED_OPERATING_SYSTEMS

    def get_configurable_network_interface(self):
        explicit_interface = configuration.get_network_interface()
        if explicit_interface:
            return explicit_interface

        interfaces = set(get_interfaces()) - {'lo'}
        interfaces = filter(lambda x: ':' not in x, interfaces)
        if len(interfaces) == 1:
            return interfaces[0]

    def backup_network_configuration(self):
        if os.path.isfile(self.interfaces_config_path):
            directory = os.path.dirname(self.interfaces_config_path)
            filename = BACKUP_FILENAME_TEMPLATE.format(
                os.path.basename(self.interfaces_config_path))
            backup_fullpath = os.path.join(directory, filename)
            if not os.path.isfile(backup_fullpath):
                shutil.copy(self.interfaces_config_path, backup_fullpath)

    def set_network(self, interface, template_path, context=None, update_network_delay_seconds=5):
        if not self.is_operating_system_supported():
            raise NotImplementedError('This operating system is not supported')

        self.read_network_configuration()

        template_text = get_package_file_content('packy_agent.configuration', template_path)

        template = Template(template_text)

        context = context or {}
        context['interface'] = interface
        content = '\n\n' + template.render(**context) + '\n\n'

        stream = BytesIO()

        written = False
        for lines in self.stanzas:
            if lines:
                first_line = lines[0]
                match = AUTO_STANZA_RE.match(first_line)
                if match:
                    stanza_iface = match.group('iface')
                    if stanza_iface == interface:
                        continue
                    else:
                        parts = stanza_iface.split(':')
                        if parts[0] == interface:
                            continue

                match = IFACE_STANZA_RE.match(first_line)
                if match:
                    stanza_iface = match.group('iface')
                    if stanza_iface == interface:
                        stream.write(content.encode('utf-8'))
                        written = True
                        continue
                    else:
                        parts = stanza_iface.split(':')
                        if parts[0] == interface:
                            # Skip virtual interfaces
                            continue

            for line in lines:
                stream.writelines(line)

        if not written:
            stream.write(content.encode('utf-8'))

        stream.seek(0)
        if configuration.is_network_configuration_enabled():
            self.backup_network_configuration()

            with atomic_write(self.interfaces_config_path, overwrite=True) as f:
                f.write(stream.read())
            run_shell_command_async('reboot now', delay_seconds=update_network_delay_seconds)
            # TODO(dmu) MEDIUM: Improve to change IP address without reboot (this includes proper restart
            #                   Packy Agent and Control Server to use update network settings
            # update_network(interface, delay_seconds=self.update_network_delay_seconds)
        else:
            logger.info('Network configuration change was disabled (for developer protection)')
            print 'Generated configuration file content:\n{}'.format(stream.read().decode('utf-8'))

    def set_static_ip_address(self, interface, ip_address, subnet_mask, default_gateway,
                              name_servers=('8.8.8.8', '8.8.4.4'), update_network_delay_seconds=5):
        context = {
            'ip_address': ip_address,
            'subnet_mask': subnet_mask,
            'default_gateway': default_gateway,
            'name_servers': name_servers,
        }
        self.set_network(interface, get_operating_system_template_path(STATIC_TEMPLATE_TYPE),
                         context=context, update_network_delay_seconds=update_network_delay_seconds)

    def set_dhcp(self, interface, update_network_delay_seconds=5):
        self.set_network(interface, get_operating_system_template_path(DHCP_TEMPLATE_TYPE),
                         update_network_delay_seconds=update_network_delay_seconds)


network_manager = NetworkManager()



from __future__ import absolute_import

import socket
import time
import logging
import select

from ryu.lib.packet.ipv4 import ipv4
from ryu.lib.packet.icmp import ICMP_TIME_EXCEEDED, ICMP_DEST_UNREACH, ICMP_PORT_UNREACH_CODE
from ryu.lib.packet.packet import Packet

from contextlib import closing
from packy_agent.modules.base.socket import TIME_SIZE_BYTES, encode_timestamp, decode_timestamp
from packy_agent.modules.base.constants import (
    IP_PACKET_HEADER_SIZE_BYTES, UDP_PACKET_HEADER_SIZE_BYTES)
from packy_agent.utils.misc import generate_random_string
from packy_agent.modules.traceroute.methods.base import guess_reply_hop_number


PACKET_HEADERS_SIZE_BYTES = IP_PACKET_HEADER_SIZE_BYTES + UDP_PACKET_HEADER_SIZE_BYTES

logger = logging.getLogger(__name__)


def send_udp_probe(raw_socket, destination_ip_address, port, ttl, packet_size_bytes=60,
                   payload=None):
    raw_socket.setsockopt(socket.SOL_IP, socket.IP_TTL, ttl)

    sent_time = time.time()
    time_part = encode_timestamp(sent_time)
    if payload is None:
        if packet_size_bytes < PACKET_HEADERS_SIZE_BYTES:
            raise ValueError('payload_size_bytes must be greater or equal to {}'.format(
                PACKET_HEADERS_SIZE_BYTES))

        payload_size_bytes = packet_size_bytes - PACKET_HEADERS_SIZE_BYTES

        if payload_size_bytes < TIME_SIZE_BYTES:
            payload = generate_random_string(payload_size_bytes)
        else:
            payload = time_part + generate_random_string(payload_size_bytes - TIME_SIZE_BYTES)
    else:
        payload = time_part + payload

    raw_socket.sendto(payload, (destination_ip_address, port))
    logger.debug('Sent UDP packet to {}:{} (ttl:{}), payload: {}'.format(
        destination_ip_address, port, ttl, payload.encode('hex')))
    return sent_time, payload


def receive_icmp_reply_basic(raw_socket, recieve_bytes=2048):
    ip_packet_data, (ip_address, _) = raw_socket.recvfrom(recieve_bytes)

    packet = Packet(ip_packet_data, parse_cls=ipv4)

    ip_packet = packet.protocols[0]
    icmp_packet = packet.protocols[1]
    logger.debug('Got ICMP packet IP checksum:{} type:{} code:{} from {}'.format(
        packet.protocols[0].csum, icmp_packet.type, icmp_packet.code, ip_address))
    if icmp_packet.type == ICMP_TIME_EXCEEDED or (icmp_packet.type == ICMP_DEST_UNREACH and
                                                  icmp_packet.code == ICMP_PORT_UNREACH_CODE):
        inner_packet = Packet(icmp_packet.data.data, parse_cls=ipv4)
        udp_packet = inner_packet.protocols[1]
        if len(inner_packet.protocols) > 2:
            udp_payload = inner_packet.protocols[2]
        else:
            udp_payload = None

        if icmp_packet.type == ICMP_TIME_EXCEEDED:
            is_destination_reached = False
        elif icmp_packet.type == ICMP_DEST_UNREACH:
            # weird, but it means that destination is reached if we get a destination unreachable
            # ICMP packet with code ICMP_PORT_UNREACH_CODE
            is_destination_reached = True
        else:
            assert False, 'Should never get here'
            raise ValueError('Unsupported ICMP packet type: {} (should never get here)'.format(
                icmp_packet.type))

        return (ip_address, ip_packet, icmp_packet, udp_packet, udp_payload,
                is_destination_reached, True)
    else:
        return ip_address, ip_packet, icmp_packet, None, None, None, False


def does_payload_match(payload, udp_payload, icmp_packet, ip_address):
    payload_len = len(payload)
    if udp_payload:
        udp_payload_len = len(udp_payload)
        if udp_payload_len != payload_len:
            logger.warning('Received payload length ({} bytes) differ from expected (sent) '
                           'length ({} bytes), but we are forgiving'.format(
                udp_payload_len, payload_len))
            min_len = min(udp_payload_len, payload_len)
            if udp_payload[:min_len] != payload[:min_len]:
                logger.debug('Discarded ICMP packet type:{} code:{} from {}: '
                             'payload prefix differs'.format(icmp_packet.type,
                                                             icmp_packet.code, ip_address))
                return False
        elif udp_payload != payload:
            logger.debug('Discarded ICMP packet type:{} code:{} from {}: '
                         'payload differs'.format(icmp_packet.type, icmp_packet.code,
                                                  ip_address))
            return False

    return True


def get_sent_time_from_payload(udp_payload):
    if udp_payload and len(udp_payload) >= TIME_SIZE_BYTES:
        return decode_timestamp(udp_payload[:TIME_SIZE_BYTES])


def receive_icmp_reply(raw_socket, port, payload, timeout, sent_time):
    payload_len = len(payload)

    time_left = timeout
    while time_left > 0:
        start_time = time.time()
        ready_sockets = select.select((raw_socket,), (), (), time_left)
        finish_time = time.time()  # because by this time the entire packet is already in OS buffer
        time_left -= (finish_time - start_time)

        if not ready_sockets[0]:  # timeout
            break

        (ip_address, ip_packet, icmp_packet, udp_packet, udp_payload, is_destination_reached,
         is_probe_response) = receive_icmp_reply_basic(raw_socket, recieve_bytes=1024 + payload_len)

        if not is_probe_response:
            logger.debug('Discarded ICMP packet type:{} code:{} from {}'.format(
                icmp_packet.type, icmp_packet.code, ip_address))
            continue

        if udp_packet.dst_port != port:
            logger.debug('Discarded ICMP packet type:{} code:{} from {} '
                         '(UDP destination port {} does not match with original {})'.format(
                icmp_packet.type, icmp_packet.code, ip_address, udp_packet.dst_port, port))
            continue

        if not does_payload_match(payload, udp_payload, icmp_packet, ip_address):
            continue

        sent_time_from_payload = get_sent_time_from_payload(udp_payload)
        if sent_time_from_payload is not None:
            sent_time = sent_time_from_payload

        from packy_agent.modules.traceroute.base import ProbeResult
        return ProbeResult(
            hop_number=None,
            reply_hop_number=guess_reply_hop_number(ip_packet.ttl),
            probe_number=None,
            hop_ip_address=ip_address,
            rtt_seconds=finish_time - sent_time,
            is_destination_reached=is_destination_reached)


def trace_hop_with_udp(destination_ip_address, port, timeout, ttl, packet_size_bytes=60):
    udp_protocol = socket.getprotobyname('udp')
    icmp_protocol = socket.getprotobyname('icmp')

    if callable(port):
        port = port()

    try:
        with closing(socket.socket(socket.AF_INET, socket.SOCK_DGRAM, udp_protocol)) as udp_socket:
            with closing(socket.socket(socket.AF_INET, socket.SOCK_RAW, icmp_protocol)) as icmp_socket:
                sent_time, payload = send_udp_probe(udp_socket, destination_ip_address, port, ttl,
                                                    packet_size_bytes)
                return receive_icmp_reply(icmp_socket, port, payload, timeout, sent_time=sent_time)
    except socket.error as (errno, msg):
        if errno == 1:  # Operation not permitted
            # TODO(dmu) HIGH: Fix to receive ICMP packet without being root
            msg += ' - process should be running as root.'
            raise socket.error(msg)

        raise



import logging

from packy_agent.modules.traceroute.methods.udp import (receive_icmp_reply_basic,
                                                        does_payload_match,
                                                        get_sent_time_from_payload)
from packy_agent.modules.traceroute.methods.base import guess_reply_hop_number

logger = logging.getLogger(__name__)


def process_udp_probe(sent_probes, icmp_socket, finish_time):
    (ip_address, ip_packet, icmp_packet, udp_packet, udp_payload,
     is_destination_reached, is_probe_response) = receive_icmp_reply_basic(icmp_socket)
    if not is_probe_response:
        logger.debug('Discarded ICMP packet type:{} code:{} from {}'.format(
            icmp_packet.type, icmp_packet.code, ip_address))
        return

    probe_port = getattr(udp_packet, 'dst_port', None)
    sent_probe = sent_probes.pop(probe_port, None)
    if not sent_probe:
        logger.debug('Discarded ICMP packet type:{} code:{} from {} '
                     '(UDP destination port {})'.format(
                      icmp_packet.type, icmp_packet.code, ip_address, probe_port))
        return

    if not does_payload_match(sent_probe.payload, udp_payload, icmp_packet, ip_address):
        return

    sent_time_from_payload = get_sent_time_from_payload(udp_payload)
    if sent_time_from_payload is None:
        actual_sent_time = sent_probe.sent_time
    else:
        actual_sent_time = sent_time_from_payload

    from packy_agent.modules.traceroute.base import ProbeResult
    return ProbeResult(
        hop_number=sent_probe.ttl,
        reply_hop_number=guess_reply_hop_number(ip_packet.ttl),
        probe_number=sent_probe.probe_number,
        hop_ip_address=ip_address,
        rtt_seconds=finish_time - actual_sent_time,
        is_destination_reached=is_destination_reached)



import logging

import flask
from flask.views import MethodView

from packy_agent.clients.packy_server import PackyServerClient
from packy_agent.control_server.forms.login import LoginForm
from packy_agent.configuration.control_server.base import (
    configuration as control_server_configuration)
from packy_agent.configuration.agent.base import configuration as agent_configuration

from packy_agent.utils.auth import set_authentication_cookie, activation_required


logger = logging.getLogger(__name__)


class LoginView(MethodView):

    @activation_required
    def get(self):
        return flask.render_template('login.html', form=LoginForm())

    @activation_required
    def post(self):
        form = LoginForm()
        if form.validate():
            # TODO(dmu) HIGH: Simplify this. Just get final config in one request
            client = PackyServerClient(
                base_url=control_server_configuration.get_server_base_url())

            if not client.validate_auth(form.email.data, form.password.data,
                                        agent_configuration.get_agent_id()):
                flask.flash('Not authenticated (invalid credentials)')
                return flask.redirect(flask.url_for('login_failure'))

            set_authentication_cookie()
            url = flask.request.args.get('next')
            if not url:
                url = flask.url_for('index')

            return flask.redirect(url)

        return flask.render_template('login.html', form=form)


class LoginFailureView(MethodView):

    def get(self):
        return flask.render_template('login_failure.html')



import logging
import re
import signal

from io import BytesIO
from subprocess import call


import supervisor.supervisorctl
from supervisor.options import ClientOptions

from packy_agent.configuration.control_server.base import configuration as control_server_configuration
from packy_agent.configuration.agent.base import configuration as agent_configuration
from packy_agent.utils.services.systemd import nginx_service
# TODO(dmu) MEDIUM: Use common code to restart supervisor from `packy-server`
from packy_agent.utils.misc import (run_shell_command_async, get_executable_path,
                                    is_inside_docker_container)

AGENT_SUPERVISOR_NAME = 'agent'
CONTROL_SERVER_SUPERVISOR_NAME = 'control-server'
WATCHDOG_SUPERVISOR_NAME = 'watchdog'
NGINX_SUPERVISOR_NAME = 'nginx'

AGENT_STATUS_RE = re.compile('^' + AGENT_SUPERVISOR_NAME + r' +([A-Z]+)(.*)$')
AGENT_UPTIME_RE = re.compile('^ +pid \d+, uptime (.*)$')

logger = logging.getLogger(__name__)


def supervisor_main(args=None, options=None, stdout=None):
    if options is None:
        options = ClientOptions()

    options.realize(args, doc=__doc__)
    c = supervisor.supervisorctl.Controller(options, stdout=stdout)

    if options.args:
        c.onecmd(' '.join(options.args))

    if options.interactive:
        c.exec_cmdloop(args, options)


def run_supervisor_command_sync(command, *args):
    logger.debug('Running sync supervisor command: %s %s', command, args)
    stdout = BytesIO()
    supervisor_main(
        ['-c', control_server_configuration.get_supervisor_configuration_file(), command] + list(args),
        stdout=stdout)

    stdout.seek(0)
    return stdout.read().decode('utf-8')


def get_supervisor_shell_command(command, args):
    args = (command,) + args
    return '{} -c {} {}'.format(get_executable_path('supervisorctl'),
        control_server_configuration.get_supervisor_configuration_file(), ' '.join(args))


def get_control_server_restart_shell_command():
    try:
        with open(control_server_configuration.get_uwsgi_pid_file()) as f:
            uwsgi_pid = int(f.read())
    except Exception:
        logger.warning('Unable to get uWSGI pid')
        return

    return 'kill -{} {}'.format(signal.SIGHUP, uwsgi_pid)


def run_supervisor_command_async(command, args, delay_seconds=None, dev_null=False):
    logger.debug('Running async supervisor command: %s %s', command, args)
    shell_command = get_supervisor_shell_command(command, args)
    run_shell_command_async(shell_command, delay_seconds, dev_null=dev_null)


def run_supervisor_command(command, args, delay_seconds=None, async=False, dev_null=False):
    if delay_seconds is None and not async:
        return run_supervisor_command_sync(command, *args)
    else:
        return run_supervisor_command_async(command, args, delay_seconds=delay_seconds,
                                            dev_null=dev_null)


def restart_component(component, delay_seconds=None, async=False, dev_null=False):
    logger.debug('Restarting {}...'.format(component))
    run_supervisor_command('restart', (component,), delay_seconds=delay_seconds,
                           async=async, dev_null=dev_null)


class ControlManager(object):

    def reboot(self, delay_seconds=None):
        agent_configuration.set_stopped(False)
        if not control_server_configuration.is_reboot_enabled():
            logger.info('Reboot was disabled (for development environment protection)')
            return

        if is_inside_docker_container():
            command = ['killall', 'supervisord']
        else:
            command = ['reboot', 'now']

        if delay_seconds is None:
            call(command)
        else:
            run_shell_command_async(' '.join(command), delay_seconds=delay_seconds)

    def restart_packy_agent(self, delay_seconds=None, dev_null=False):
        # TODO(dmu) HIGH: Restart Packy Agent asynchronously?
        # TODO(dmu) MEDIUM: Use common code to restart supervisor from `packy-server`
        agent_configuration.set_stopped(False)
        restart_component(AGENT_SUPERVISOR_NAME, delay_seconds=delay_seconds, dev_null=dev_null)

    def restart_watchdog(self, delay_seconds=None):
        restart_component(WATCHDOG_SUPERVISOR_NAME, delay_seconds=delay_seconds)

    def restart_nginx(self, delay_seconds=None):
        if is_inside_docker_container():
            restart_component(NGINX_SUPERVISOR_NAME, delay_seconds=delay_seconds)
        else:
            nginx_service.restart()

    def start_packy_agent(self):
        logger.debug('Starting agent...')
        agent_configuration.set_stopped(False)
        run_supervisor_command_sync('start', AGENT_SUPERVISOR_NAME)

    def stop_packy_agent(self):
        logger.debug('Stopping agent...')
        agent_configuration.set_stopped(True)
        run_supervisor_command_sync('stop', AGENT_SUPERVISOR_NAME)

    def restart_control_server(self, delay_seconds=None):
        logger.debug('Restarting control server...')
        command = get_control_server_restart_shell_command()
        if command:
            run_shell_command_async(command, delay_seconds=delay_seconds)
        # TODO(dmu) MEDIUM: `uwsgi.reload()` breaks current request and we do not manage to send
        #                   response to user
        # import uwsgi
        # uwsgi.reload()
        # TODO(dmu) MEDIUM: Restart via supervisor does not work properly under Docker, because
        #                   supervisord under Docker kills background subprocess that runs
        #                   restart procedure (maybe it is Alpine specific)
        # run_supervisor_command('restart', (CONTROL_SERVER_SUPERVISOR_NAME,),
        #                        delay_seconds=delay_seconds)

    def restart_all(self, delay_seconds=None):
        """Restart all components"""
        logger.debug('Restarting all components...')
        self.restart_packy_agent()
        self.restart_watchdog()
        self.restart_nginx()
        self.restart_control_server(delay_seconds=delay_seconds)
        # TODO(dmu) MEDIUM: Restart via supervisor does not work properly under Docker, because
        #                   supervisord under Docker kills background subprocess that runs
        #                   restart procedure (maybe it is Alpine specific)
        # run_supervisor_command('restart', ('all',), delay_seconds=delay_seconds)

    def get_packy_agent_status(self):
        status = None
        uptime = None

        output = run_supervisor_command_sync('status', AGENT_SUPERVISOR_NAME)
        output = output.rstrip()
        match_status = AGENT_STATUS_RE.match(output)
        if match_status:
            status, tail = match_status.groups()
            match_uptime = AGENT_UPTIME_RE.match(tail)
            if match_uptime:
                 uptime = match_uptime.group(1)

        return status, uptime


control_manager = ControlManager()



import logging

import flask
from raven.contrib.flask import Sentry

import packy_agent
from packy_agent.utils.pkg_resources import get_package_file_full_path

from packy_agent.control_server.views.index import IndexView
from packy_agent.control_server.views.login import LoginView, LoginFailureView
from packy_agent.control_server.views.logout import LogoutView
from packy_agent.control_server.views.success import SuccessView
from packy_agent.control_server.views.failure import FailureView
from packy_agent.control_server.views.network import NetworkView
from packy_agent.control_server.views.control import ControlView
from packy_agent.control_server.views.upgrade import UpgradeView
from packy_agent.control_server.views.reset import ResetView
from packy_agent.control_server.views.debug import DebugView
from packy_agent.control_server.views.activate import ActivateView, ActivationFailureView
from packy_agent.control_server.views import errors as error_handlers
from packy_agent.utils.misc import (
    is_inside_docker_container, dump_version, remove_upgrade_in_progress_lock,
    is_upgrade_in_progress)
from packy_agent.utils.sentry import get_raven_client

from packy_agent.configuration.control_server.base import configuration


def configure(app):
    # TODO(dmu) LOW: Refactor to make autooverride from options if there are more options to
    #                override
    app.config.from_mapping(configuration.get_flask_configuration())


def setup_routes(app):
    app.add_url_rule('/', view_func=IndexView.as_view('index'))
    app.add_url_rule('/network/', view_func=NetworkView.as_view('network'))
    app.add_url_rule('/control/', view_func=ControlView.as_view('control'))
    app.add_url_rule('/upgrade/', view_func=UpgradeView.as_view('upgrade'))
    app.add_url_rule('/reset/', view_func=ResetView.as_view('reset'))
    app.add_url_rule('/debug/', view_func=DebugView.as_view('debug'))
    app.add_url_rule('/success/', view_func=SuccessView.as_view('success'))
    app.add_url_rule('/failure/', view_func=FailureView.as_view('failure'))
    app.add_url_rule('/activate/', view_func=ActivateView.as_view('activate'))
    app.add_url_rule('/activation_failure/',
                     view_func=ActivationFailureView.as_view('activation_failure'))
    app.add_url_rule('/login/', view_func=LoginView.as_view('login'))
    app.add_url_rule('/login_failure/', view_func=LoginFailureView.as_view('login_failure'))
    app.add_url_rule('/logout/', view_func=LogoutView.as_view('logout'))


def setup_context_processors(app):

    @app.context_processor
    def extra_context_processor():
        return {
            'version': packy_agent.__version__,
            'is_inside_docker_container': is_inside_docker_container(),
            'is_debug_mode': configuration.is_debug_mode(),
            'configuration': configuration,
            'is_upgrade_in_progress': is_upgrade_in_progress,
        }


def setup_error_handlers(app):
    app.register_error_handler(500, error_handlers.handle_http500)
    app.register_error_handler(401, error_handlers.handle_http401)


def get_app():
    raven_client = get_raven_client('control server')
    # TODO(dmu) HIGH: Improve logging configuration
    logging.basicConfig(format='%(asctime)-15s %(levelname)s [%(name)s]: %(message)s',
                        level=logging.DEBUG)

    app = flask.Flask(__name__,
                      template_folder=get_package_file_full_path(__name__, 'templates'),
                      static_folder=get_package_file_full_path(__name__, 'static'),
                      static_url_path='/assets')
    Sentry(app, client=raven_client)

    configure(app)
    setup_routes(app)
    setup_context_processors(app)
    setup_error_handlers(app)

    remove_upgrade_in_progress_lock()
    dump_version('/etc/packy-control-server-version', 'PACKY_CONTROL_SERVER_VERSION')

    return app



import logging

from celery import shared_task

from packy_agent.configuration.agent.base import configuration
from packy_agent.managers.control import control_manager

logger = logging.getLogger(__name__)


@shared_task()
def update_config():
    configuration.update_local_configuration_from_server()
    control_manager.restart_control_server()
    control_manager.restart_watchdog()



import flask
from flask.views import MethodView

from packy_agent.control_server.forms.network import NetworkForm
from packy_agent.configuration.control_server.base import configuration
from packy_agent.managers.network import network_manager
from packy_agent.control_server.views.base import smart_redirect
from packy_agent.utils.auth import activation_and_authentication_required


DELAY_SECONDS = 5


class NetworkView(MethodView):

    def get_form_context(self):
        is_operating_system_supported = network_manager.is_operating_system_supported()
        network_interface = network_manager.get_configurable_network_interface()

        if is_operating_system_supported and network_interface:
            form_kwargs = network_manager.get_current_configuration(network_interface)
            form_kwargs['network_interface'] = network_interface
            form = NetworkForm(**form_kwargs)
        else:
            form = None

        context = {
            'form': form,
            'is_network_configuration_enabled': configuration.is_network_configuration_enabled(),
            'is_operating_system_supported': is_operating_system_supported,
            'network_interface': network_interface,
            'active_menu_item': 'network',
        }
        return context

    @activation_and_authentication_required
    def get(self):
        context = self.get_form_context()
        return flask.render_template('network.html', **context)

    @activation_and_authentication_required
    def post(self):
        form = NetworkForm()
        if form.validate():
            if form.dhcp.data:
                network_manager.set_dhcp(interface=form.network_interface.data)
            else:
                kwargs = dict(
                    ip_address=form.ip_address.data,
                    subnet_mask=form.subnet_mask.data,
                    default_gateway=form.default_gateway.data,
                    interface=form.network_interface.data
                )
                if form.name_servers.data:
                    kwargs['name_servers'] = form.name_servers.data.split(',')
                network_manager.set_static_ip_address(**kwargs)

            if configuration.is_network_configuration_enabled():
                flask.flash('Operation will start in {} seconds...'.format(DELAY_SECONDS))
            else:
                flask.flash('Please, see logs for more information')

            return smart_redirect('success', 'network')

        context = self.get_form_context()
        context['form'] = form
        return flask.render_template('network.html', **context)



from __future__ import absolute_import
import json


from schematics import Model


class CustomModel(Model):

    def to_json(self, *args, **kwargs):
        return json.dumps(self.to_primitive(*args, **kwargs))



from __future__ import print_function

import sys

from celery import Celery
from raven.contrib.celery import register_signal, register_logger_signal

from packy_agent.configuration.agent.base import configuration
from packy_agent.utils.misc import dump_version, remove_upgrade_in_progress_lock
from packy_agent.utils.sentry import get_raven_client


class CustomCelery(Celery):

    def __init__(self, *args, **kwargs):
        self.raven_client = kwargs.pop('raven_client', None)
        super(CustomCelery, self).__init__(*args, **kwargs)

    def gen_task_name(self, name, module):
        if self.main:
            return self.main + '.' + name

        return super(CustomCelery, self).gen_task_name(name, module)

    def on_configure(self):
        if self.raven_client:
            register_logger_signal(self.raven_client)
            register_signal(self.raven_client)


def get_celery_app(raven_client=None):
    agent_id = configuration.get_agent_id()
    if agent_id is None:
        return

    configuration.update_local_configuration_from_server()

    application = CustomCelery(configuration.get_tasks_prefix(), raven_client=raven_client)
    application.config_from_object(configuration.get_celery_configuration())

    __import__('packy_agent.tasks')  # register tasks

    return application


# Just creation of Raven Client logs to Sentry in case exception
app = get_celery_app(raven_client=get_raven_client('agent'))
remove_upgrade_in_progress_lock()
dump_version('/etc/packy-agent-version', 'PACKY_AGENT_VERSION')
if not app:
    # TODO(dmu) LOW: Provide a more elegant way to exit Celery until Agent is activated
    print('Packy Agent ID (packy.agent.agent_id) is not known (maybe agent has not '
          'been activated yet), will exit', file=sys.stderr)
    sys.exit(1)



import os
import stat

from packy_agent.utils.services.base import ServiceBase
from packy_agent.utils.shell import run_command
from packy_agent.utils.output import write_to_console_or_file


class SystemdService(ServiceBase):

    RELOAD_SYSTEMD_CONFIGURATION_COMMAND = 'systemctl daemon-reload'
    MAKE_START_ON_REBOOT_COMMAND_TEMPLATE = 'update-rc.d {name} defaults'
    ENABLE_UNIT_SERVICE_COMMAND_TEMPLATE = 'systemctl enable {name}.service'
    IS_ACTIVE_SERVICE_COMMAND_TEMPLATE = 'systemctl is-active {name}.service'

    START_COMMAND_TEMPLATE = 'service {name} start'
    STOP_COMMAND_TEMPLATE = 'service {name} stop'
    RESTART_COMMAND_TEMPLATE = 'service {name} restart'
    RELOAD_COMMAND_TEMPLATE = 'service {name} reload'

    SYSTEMCTL_START_COMMAND_TEMPLATE = 'systemctl start {name}.service'
    SYSTEMCTL_STOP_COMMAND_TEMPLATE = 'systemctl stop {name}.service'
    SYSTEMCTL_RESTART_COMMAND_TEMPLATE = 'systemctl restart {name}.service'
    SYSTEMCTL_RELOAD_COMMAND_TEMPLATE = 'systemctl reload {name}.service'

    INITD_PATH_TEMPLATE = '/etc/init.d/{name}'
    # TODO(dmu) MEDIUM: Is `UNIT_SERVICE_PATH_TEMPLATE` different on OSes other than Ubuntu
    UNIT_SERVICE_PATH_TEMPLATE = '/lib/systemd/system/{name}.service'

    def __init__(self, name):
        super(SystemdService, self).__init__(name)

        self.start_command = self.START_COMMAND_TEMPLATE.format(name=name)
        self.stop_command = self.STOP_COMMAND_TEMPLATE.format(name=name)
        self.restart_command = self.RESTART_COMMAND_TEMPLATE.format(name=name)
        self.reload_command = self.RELOAD_COMMAND_TEMPLATE.format(name=name)

        self.systemctl_start_command = self.SYSTEMCTL_START_COMMAND_TEMPLATE.format(name=name)
        self.systemctl_stop_command = self.SYSTEMCTL_STOP_COMMAND_TEMPLATE.format(name=name)
        self.systemctl_restart_command = self.SYSTEMCTL_RESTART_COMMAND_TEMPLATE.format(name=name)
        self.systemctl_reload_command = self.SYSTEMCTL_RELOAD_COMMAND_TEMPLATE.format(name=name)

        self.make_start_on_reboot_command = self.MAKE_START_ON_REBOOT_COMMAND_TEMPLATE.format(
            name=name)
        self.enable_unit_service_command = self.ENABLE_UNIT_SERVICE_COMMAND_TEMPLATE.format(
            name=name)
        self.is_active_service_command = self.IS_ACTIVE_SERVICE_COMMAND_TEMPLATE.format(name=name)

        self.initd_path = self.INITD_PATH_TEMPLATE.format(name=name)
        self.unit_service_path = self.UNIT_SERVICE_PATH_TEMPLATE.format(name=name)

    @classmethod
    def reload_systemd_configuration(cls, raise_exception=False):
        run_command(cls.RELOAD_SYSTEMD_CONFIGURATION_COMMAND, raise_exception=raise_exception)

    def start(self, systemctl=False, raise_exception=False):
        command = self.systemctl_start_command if systemctl else self.start_command
        run_command(command, raise_exception=raise_exception)

    def stop(self, systemctl=False, raise_exception=False):
        command = self.systemctl_stop_command if systemctl else self.stop_command
        run_command(command, raise_exception=raise_exception)

    def restart(self, systemctl=False, raise_exception=False):
        command = self.systemctl_restart_command if systemctl else self.restart_command
        run_command(command, raise_exception=raise_exception)

    def reload(self, systemctl=False, raise_exception=False):
        command = self.systemctl_reload_command if systemctl else self.reload_command
        run_command(command, raise_exception=raise_exception)

    def make_start_on_reboot(self, raise_exception=False):
        run_command(self.make_start_on_reboot_command, raise_exception=raise_exception)

    def enable_unit_service(self, raise_exception=False):
        run_command(self.enable_unit_service_command, raise_exception=raise_exception)

    def update_initd_script(self, content):
        initd_path = self.initd_path
        write_to_console_or_file(initd_path, content)
        os.chmod(initd_path, os.stat(initd_path).st_mode | stat.S_IEXEC)
        self.reload_systemd_configuration(raise_exception=True)
        self.make_start_on_reboot(raise_exception=True)

    def update_unit_service_configuration(self, content):
        write_to_console_or_file(self.unit_service_path, content)
        self.enable_unit_service(raise_exception=True)
        self.reload_systemd_configuration(raise_exception=True)
        self.restart(systemctl=True, raise_exception=True)

    def is_active(self, raise_exception=False):
        output = run_command(self.is_active_service_command, raise_exception=raise_exception)
        return output == 'active'


packy_service = SystemdService('packy')
nginx_service = SystemdService('nginx')
log2ram_service = SystemdService('log2ram')



import os.path
import logging
import yaml
import pprint
import tailer

import flask
from flask.views import MethodView

from packy_agent.exceptions import ImproperlyConfigured
from packy_agent.configuration.agent.base import configuration as agent_configuration
from packy_agent.configuration.control_server.base import configuration as control_server_configuration
from packy_agent.utils.auth import activation_and_authentication_required

TAIL_SIZE = 100


logger = logging.getLogger(__name__)


def tail(filename, size=TAIL_SIZE, safe=True):
    try:
        with open(filename) as f:
            return '\n'.join(tailer.tail(f, size))
    except Exception:
        if not safe:
            raise


class DebugView(MethodView):

    @activation_and_authentication_required
    def get(self):

        logs = (
            ('Supervisor log',
             control_server_configuration.get_supervisor_log_file(),
             tail(control_server_configuration.get_supervisor_log_file())),
            ('Packy Agent stdout log',
             control_server_configuration.get_agent_stdout_log_file(),
             tail(control_server_configuration.get_agent_stdout_log_file())),
            ('Packy Agent stderr log',
             control_server_configuration.get_agent_stderr_log_file(),
             tail(control_server_configuration.get_agent_stderr_log_file())),
            ('Control Server stdout log',
             control_server_configuration.get_control_server_stdout_log_file(),
             tail(control_server_configuration.get_control_server_stdout_log_file())),
            ('Control Server stderr log',
             control_server_configuration.get_control_server_stderr_log_file(),
             tail(control_server_configuration.get_control_server_stderr_log_file())),
        )

        packy_agent_configuration_file = agent_configuration.get_local_configuration_file_path()
        if os.path.isfile(packy_agent_configuration_file):
            with open(packy_agent_configuration_file) as f:
                packy_agent_configuration_file_content = f.read()
        else:
            packy_agent_configuration_file += ' (does not exist)'
            packy_agent_configuration_file_content = ''

        packy_agent_custom_configuration = yaml.safe_dump(
            agent_configuration.get_packy_agent_configuration(), default_flow_style=False)
        try:
            packy_agent_celery_configuration = pprint.pformat(
                    agent_configuration.get_celery_configuration(), indent=4)
        except ImproperlyConfigured:
            packy_agent_celery_configuration = '(not configured)'

        control_server_configuration_file = (
            control_server_configuration.get_local_configuration_file_path())
        if os.path.isfile(control_server_configuration_file):
            with open(control_server_configuration_file) as f:
                control_server_configuration_file_content = f.read()
        else:
            control_server_configuration_file += ' (does not exist)'
            control_server_configuration_file_content = ''

        control_server_custom_configuration = yaml.safe_dump(
            control_server_configuration.get_control_server_configuration(),
            default_flow_style=False)
        control_server_flask_configuration = pprint.pformat(
            control_server_configuration.get_flask_configuration(), indent=4)

        context = {
            'packy_agent_configuration_file': packy_agent_configuration_file,
            'packy_agent_configuration_file_content': packy_agent_configuration_file_content,
            'packy_agent_custom_configuration': packy_agent_custom_configuration,
            'packy_agent_celery_configuration': packy_agent_celery_configuration,
            'control_server_configuration_file': control_server_configuration_file,
            'control_server_configuration_file_content': control_server_configuration_file_content,
            'control_server_custom_configuration': control_server_custom_configuration,
            'control_server_celery_configuration': control_server_flask_configuration,
            'logs': logs,
            'active_menu_item': 'debug',
        }

        return flask.render_template('debug.html', **context)



import logging
import json

from celery import shared_task

import psutil

import packy_agent
from packy_agent.clients.packy_server import packy_server_client_factory
from packy_agent.configuration.agent.base import configuration
from packy_agent.managers.network import network_manager
from packy_agent.utils.misc import atomic_write

logger = logging.getLogger(__name__)


def load_data_usage():
    filename = configuration.get_data_usage_file()
    try:
        with open(filename, 'r') as f:
            content = f.read()
            if content:
                content_json = json.loads(content)
                return content_json.get('bytes_sent'), content_json.get('bytes_received')
    except IOError:
        pass

    return None, None


def set_data_usage(bytes_sent, bytes_received):
    with atomic_write(configuration.get_data_usage_file(), overwrite=True) as f:
        f.write(json.dumps({'bytes_sent': bytes_sent, 'bytes_received': bytes_received}))


def get_data_usage():
    snetio = psutil.net_io_counters(pernic=True, nowrap=True)
    snetio.pop('lo', None)

    bytes_sent = 0
    bytes_received = 0

    for value in snetio.itervalues():
        bytes_sent += value.bytes_sent
        bytes_received += value.bytes_recv

    return bytes_sent, bytes_received


@shared_task()
def heartbeat():
    if not configuration.is_heartbeat_enabled():
        logger.debug('Heartbeat is disabled')
        return

    prev_bytes_sent, prev_bytes_received = load_data_usage()
    bytes_sent, bytes_received = get_data_usage()
    set_data_usage(bytes_sent, bytes_received)

    client = packy_server_client_factory.get_client_by_configuration(configuration)
    try:
        client.update_status(version=packy_agent.__version__,
                             ip_address=network_manager.get_actual_ip_address(),
                             prev_bytes_sent=prev_bytes_sent,
                             prev_bytes_received=prev_bytes_received,
                             bytes_sent=bytes_sent,
                             bytes_received=bytes_received)
    except Exception:
        logger.warning('Error while sending heartbeat', exc_info=True)



import logging

import flask
from flask.views import MethodView

from packy_agent.control_server.forms.action import ActionForm
from packy_agent.configuration.control_server.base import configuration
from packy_agent.managers.control import control_manager
from packy_agent.managers.reset import reset_manager
from packy_agent.utils.misc import is_inside_docker_container
from packy_agent.utils.auth import activation_and_authentication_required, logout


OPERATION_DELAY_SECONDS = 5

logger = logging.getLogger(__name__)


def do_action(action):
    if action == 'deactivate_and_reset':
        if (not is_inside_docker_container() and
                not configuration.is_network_configuration_enabled()):
            flask.flash('Please, see logs for more information')

        reset_manager.reset_all()
        logout()
        control_manager.reboot(OPERATION_DELAY_SECONDS)
    elif action == 'deactivate':
        control_manager.stop_packy_agent()
        reset_manager.deactivate()
        logout()
    elif action == 'reset_network':
        if not configuration.is_network_configuration_enabled():
            flask.flash('Please, see logs for more information')

        reset_manager.reset_network_configuration()
        control_manager.reboot(OPERATION_DELAY_SECONDS)
    else:
        raise ValueError('Unknown action: {}'.format(action))


class ResetView(MethodView):

    @activation_and_authentication_required
    def get(self):
        form = ActionForm()
        context = {
            'form': form,
            'active_menu_item': 'reset',
        }

        return flask.render_template('reset.html', **context)

    @activation_and_authentication_required
    def post(self):
        form = ActionForm()
        if form.validate():
            action = flask.request.form['action']
            do_action(action)
            flask.flash('Operation will start in {} seconds...'.format(OPERATION_DELAY_SECONDS))
            return flask.redirect(flask.url_for('index'))

        return flask.render_template('reset.html', form=form)



import urllib

import flask


def smart_redirect(endpoint, back_endpoint, button_text='Back'):
    back = urllib.urlencode({
        'back_url': flask.url_for(back_endpoint),
        'button_text': button_text,
    })
    url = flask.url_for(endpoint) + '?' + back
    return flask.redirect(url)



import sys
import argparse

from packy_agent.managers.control import control_manager


def restart():
    control_manager.restart_all()


def entry():
    from packy_agent.cli import generate_uwsgi_ini as the_module
    parser = argparse.ArgumentParser(
        # Cannot use __name__, because when module is run as script it is equal to __main__
        prog='python -m {}'.format(the_module.__name__),
        formatter_class=argparse.ArgumentDefaultsHelpFormatter)

    args = parser.parse_args()

    return restart()


if __name__ == '__main__':
    sys.exit(entry())


import logging
import time
import base64
import socket
import uuid
from urlparse import urljoin
from urllib import urlencode

import distro
import cachetools
import requests

import packy_agent
from packy_agent.exceptions import AuthenticationError
from packy_agent.utils.collections import set_if

packy_server_client = None
logger = logging.getLogger(__name__)

DEFAULT_TIMEOUT = 15


def get_platform():
    from packy_agent.utils.misc import is_inside_docker_container
    platform = distro.name(pretty=True) or 'unidentified'
    if is_inside_docker_container():
        platform = 'Docker/{}'.format(platform)

    return platform


def make_basic_auth_headers(username, password):
    return {'Authorization': 'Basic ' + base64.b64encode('%s:%s' % (username, password))}


class PackyServerClient(object):
    # TODO(dmu) HIGH: Implement support for persistent HTTP?

    OAUTH_TOKEN_PATH = 'oauth/token/'

    PENDING_AGENT_PATH = 'api/v2/pending-agent/'
    AGENT_VERSION_PATH = 'api/v2/agent-version/'
    AGENT_TEMPLATE_PATH = 'api/v2/user/agent/'
    AGENT_INSTANCE_TEMPLATE_PATH = AGENT_TEMPLATE_PATH + '{}/'

    UPDATE_STATUS_TEMPLATE_PATH = AGENT_INSTANCE_TEMPLATE_PATH
    AGENT_CONFIG_TEMPLATE_PATH = AGENT_INSTANCE_TEMPLATE_PATH + 'config/'
    VALIDATE_AUTH_TEMPLATE_PATH = AGENT_INSTANCE_TEMPLATE_PATH + 'validate-auth/'
    HEARTBEAT_TEMPLATE_PATH = AGENT_INSTANCE_TEMPLATE_PATH + 'heartbeat/'

    PING_MEASUREMENT_TEMPLATE_PATH = AGENT_INSTANCE_TEMPLATE_PATH + 'module/ping/measurement/'
    SPEEDTEST_MEASUREMENT_TEMPLATE_PATH = AGENT_INSTANCE_TEMPLATE_PATH + 'module/speedtest/measurement/'
    TRACE_MEASUREMENT_TEMPLATE_PATH = AGENT_INSTANCE_TEMPLATE_PATH + 'module/trace/measurement/'
    HTTP_MEASUREMENT_TEMPLATE_PATH = AGENT_INSTANCE_TEMPLATE_PATH + 'module/http/measurement/'

    def __init__(self, base_url, client_id=None, client_secret=None, agent_id=None,
                 timeout=DEFAULT_TIMEOUT):
        self.base_url = base_url

        self.client_id = client_id
        self.client_secret = client_secret
        self.agent_id = agent_id
        self.timeout = timeout

        self.token = None
        self.token_type = None
        self.token_expiration = None

        self.user_agent = 'packy-agent/{} ({})'.format(packy_agent.__version__, get_platform())

        self.oauth_token_url = urljoin(base_url, self.OAUTH_TOKEN_PATH)
        self.pending_agent_url = urljoin(base_url, self.PENDING_AGENT_PATH)

        self.agent_version_url = urljoin(base_url, self.AGENT_VERSION_PATH)
        self.agent_config_template_url = urljoin(base_url, self.AGENT_CONFIG_TEMPLATE_PATH)

        self.agent_resource_url = urljoin(base_url, self.AGENT_TEMPLATE_PATH)

        self.agent_instance_template_url = urljoin(base_url, self.AGENT_INSTANCE_TEMPLATE_PATH)
        self.update_status_template_url = urljoin(base_url, self.UPDATE_STATUS_TEMPLATE_PATH)
        self.validate_auth_template_url = urljoin(base_url, self.VALIDATE_AUTH_TEMPLATE_PATH)

        self.ping_measurement_template_url = urljoin(base_url, self.PING_MEASUREMENT_TEMPLATE_PATH)
        self.speedtest_measurement_template_url = urljoin(base_url, self.SPEEDTEST_MEASUREMENT_TEMPLATE_PATH)
        self.trace_measurement_template_url = urljoin(base_url, self.TRACE_MEASUREMENT_TEMPLATE_PATH)
        self.http_measurement_template_url = urljoin(base_url, self.HTTP_MEASUREMENT_TEMPLATE_PATH)

        self.heartbeat_template_url = urljoin(base_url, self.HEARTBEAT_TEMPLATE_PATH)

    def set_client_id(self, client_id):
        self.client_id = client_id

    def set_client_secret(self, client_secret):
        self.client_secret = client_secret

    # TODO(dmu) MEDIUM: Get rid of `auto_login`. Support multiple authentication methods instead
    def request(self, url, data=None, json=None, method=None, headers=None, auto_login=True):
        if not method:
            if data is not None or json is not None:
                method = 'POST'
            else:
                method = 'GET'

        # TODO(dmu) HIGH: Implement token expiration support
        if auto_login and (not self.token or time.time() >= self.token_expiration):
            self.login()

        headers = headers.copy() if headers else {}
        if self.token_type and self.token:
            headers.setdefault('Authorization', self.token_type + ' ' + self.token)
        headers.setdefault('Content-Type', 'application/json')
        headers.setdefault('User-Agent', self.user_agent)

        tries = 2
        while tries > 0:
            tries -= 1
            if method == 'GET':
                response = requests.get(url, headers=headers, timeout=self.timeout)
            elif method == 'POST':
                response = requests.post(url, data=data, json=json, headers=headers,
                                         timeout=self.timeout)
            elif method == 'PATCH':
                response = requests.patch(url, data=data, json=json, headers=headers,
                                          timeout=self.timeout)
            else:
                raise NotImplementedError('Method {} is not supported'.format(method))

            if response.status_code == 401:
                # TODO(dmu) MEDIUM: Provide a better error response for expired token
                logger.debug('Maybe token has expired, will login again and retry: HTTP%s %s',
                             response.status_code, response.content)
                # TODO(dmu) MEDIUM: Support for exchange token or maybe switch to JWT
                self.login()
            else:
                break

        if response.status_code >= 400:
            logger.debug('HTTP%s: %s', response.status_code, response.content)

        return response

    def get_version_max(self):
        response = self.request(self.agent_version_url, auto_login=False)
        response.raise_for_status()
        return response.json().get('version_max')

    def login(self):
        url = urljoin(self.base_url, self.OAUTH_TOKEN_PATH)
        response = requests.post(url,
                                 data={'grant_type': 'client_credentials',
                                       'client_id': self.client_id,
                                       'client_secret': self.client_secret},
                                 timeout=self.timeout)
        if response.status_code == 200:
            response_json = response.json()
            self.token = response_json['access_token']
            self.token_type = response_json['token_type']
            # expire earlier
            self.token_expiration = time.time() + response_json['expires_in'] * 0.9
        else:
            raise AuthenticationError('Not authenticated: %s', response.content)

    def create_agent(self, email, password):
        payload = {
            # TODO(dmu) MEDIUM: What `key` is used for?
            # TODO(dmu) LOW: Why not UUID4?
            'key': uuid.uuid1().hex,
            'name': socket.gethostname()
        }

        return self.request(self.agent_resource_url, json=payload,
                            headers=make_basic_auth_headers(email, password), auto_login=False)

    def get_inactive_agents(self, email, password):
        params = {'fields': 'id,name', 'is_active': 'false'}
        return self.request(
            self.agent_resource_url + '?{}'.format(urlencode(params)),
            headers=make_basic_auth_headers(email, password), auto_login=False
        )

    def activate_agent(self, email, password):
        return self.request(
            self.update_status_template_url.format(self.agent_id) + '?{}'.format(
                urlencode({'with_configuration': 'yes'})),
            json={'is_active': True}, method='PATCH',
            headers=make_basic_auth_headers(email, password), auto_login=False)

    def deactivate_agent(self):
        return self.request(self.update_status_template_url.format(self.agent_id),
                            json={'is_active': False}, method='PATCH')

    def validate_auth(self, username, password, agent_id):
        response = requests.post(self.validate_auth_template_url.format(agent_id),
                                 headers=make_basic_auth_headers(username, password),
                                 timeout=self.timeout)
        status_code = response.status_code
        if status_code == 401:
            return False

        if status_code == 404:
            logger.info('HTTP404: Probably credentials of %s do not match agent_id: %s',
                        username, agent_id)
            return False

        response.raise_for_status()
        return True

    def get_agent(self):
        return self.request(self.agent_instance_template_url.format(self.agent_id))

    def get_agent_config(self, agent_id=None):
        # TODO(dmu) MEDIUM: Use only `self.agent_id`
        url = self.agent_config_template_url.format(agent_id or self.agent_id)
        return self.request(url)

    def notify_server(self, agent_control_server_url):
        # TODO(dmu) LOW: Why do not we use self.request() or refactor?
        response = requests.post(self.pending_agent_url,
                                 json={'agent_control_server_url': agent_control_server_url},
                                 timeout=self.timeout)
        response.raise_for_status()

    def submit_ping_measurement(self, data):
        return self.request(self.ping_measurement_template_url.format(self.agent_id), json=data)

    def submit_speedtest_measurement(self, data):
        return self.request(self.speedtest_measurement_template_url.format(self.agent_id),
                            json=data)

    def submit_trace_measurement(self, data):
        return self.request(self.trace_measurement_template_url.format(self.agent_id), json=data)

    def submit_http_measurement(self, data):
        return self.request(self.http_measurement_template_url.format(self.agent_id), json=data)

    def send_heartbeat(self):
        return self.request(self.heartbeat_template_url.format(self.agent_id), method='POST')

    def update_status(self, version=None, ip_address=None, is_online=True,
                      bytes_sent=None, prev_bytes_sent=None, bytes_received=None,
                      prev_bytes_received=None):
        payload = {'is_online': is_online}
        set_if(payload, 'version', version)
        set_if(payload, 'ip_address', ip_address)
        set_if(payload, 'bytes_sent', bytes_sent, if_callable=lambda x: x is not None)
        set_if(payload, 'prev_bytes_sent', prev_bytes_sent, if_callable=lambda x: x is not None)
        set_if(payload, 'bytes_received', bytes_received, if_callable=lambda x: x is not None)
        set_if(payload, 'prev_bytes_received', prev_bytes_received,
               if_callable=lambda x: x is not None)

        return self.request(self.update_status_template_url.format(self.agent_id),
                            json=payload, method='PATCH')


class PackyServerClientFactory(object):
    # This factory is needed because potentially configuration may change with configuration
    # update and we may need to login again to server

    def __init__(self):
        # We could implement differently taking in account that at the moment we need only one
        # client, but it would result in the code of the same complexity, but less flexible
        self.clients = cachetools.LRUCache(1)

    def get_client(self, base_url, client_id=None, client_secret=None, agent_id=None):
        key = (base_url, client_id, client_secret, agent_id)
        client = self.clients.get(key)

        if client:
            # It is not expected that someone modifies these attributes for factory acquired clients
            assert client.base_url == base_url
            assert client.client_id == client_id
            assert client.client_secret == client_secret
            assert client.agent_id == agent_id
        else:
            client = PackyServerClient(base_url=base_url, client_id=client_id,
                                       client_secret=client_secret, agent_id=agent_id)
            self.clients[key] = client

        return client

    def get_client_by_configuration(self, configuration):
        return self.get_client(
            base_url=configuration.get_server_base_url(),
            client_id=configuration.get_client_id(),
            client_secret=configuration.get_client_secret(),
            agent_id=configuration.get_agent_id(),
        )


packy_server_client_factory = PackyServerClientFactory()



import logging

import flask
from flask.views import MethodView

from packy_agent.configuration.agent.base import configuration as agent_configuration
from packy_agent.managers.control import control_manager
from packy_agent.utils.auth import activation_and_authentication_required

logger = logging.getLogger(__name__)


class IndexView(MethodView):

    @activation_and_authentication_required
    def get(self):
        agent_id = agent_configuration.get_agent_id()
        if agent_id is None:
            logger.debug('Agent is not activated yet, redirecting to activation')
            return flask.redirect(flask.url_for('activate'))
        else:
            agent_status = control_manager.get_packy_agent_status()
            agent_status_line = agent_status[0] or 'UNKNOWN'
            if agent_status[1]:
                agent_status_line += ' (uptime {})'.format(agent_status[1])

            context = {
                'agent_status': agent_status_line,
                'active_menu_item': 'status',
            }
            return flask.render_template('index.html', **context)



import logging
from retry.api import retry_call

from celery import shared_task

from packy_agent.clients.packy_server import packy_server_client_factory
from packy_agent.configuration.agent.base import configuration
from packy_agent.modules.http.base import get_http_measurement
from packy_agent.tasks.modules.base import run_exclusive
from packy_agent.exceptions import ImproperlyConfigured

logger = logging.getLogger(__name__)


def http_impl():
    module_config = configuration.get_http_module_configuration()
    if not module_config:
        raise ImproperlyConfigured('No configuration for `http` module found')

    urls = module_config.get('urls')
    if not urls:
        raise ImproperlyConfigured('URLs are not configured for `http` module')

    follow_redirects = module_config.get('follow_redirects', False)

    client = packy_server_client_factory.get_client_by_configuration(configuration)
    for url in urls:

        # TODO(dmu) MEDIUM: Figure out what happens inside `curl` library and remove this
        #                   workaround
        for _ in xrange(10):
            measurement = get_http_measurement(url, follow_redirects=follow_redirects)
            if measurement.total_ms < 4000000 and measurement.namelookup_ms < 4000000:
                break
        else:
            logger.error('Failed to get proper measurement from curl for {]'.format(url))
            continue

        measurement_json = measurement.to_primitive()
        retry_call(client.submit_http_measurement, (measurement_json,),
                   tries=configuration.get_submit_retries())


@shared_task
def http():
    run_exclusive(http_impl)



import os
import os.path
import yaml

from packy_agent.utils.collections import deep_update

from packy_agent.configuration import common as common_configuration
from packy_agent.configuration.base import BaseConfiguration
from packy_agent.utils.pkg_resources import get_package_file_content
from packy_agent.utils.misc import generate_random_string


def generate_flask_secret_key():
    return generate_random_string(128)


class ControlServerConfiguration(BaseConfiguration):

    FLASK_CONFIGURATION_KEY = 'flask'
    CONTROL_SERVER_CONFIGURATION_KEY = 'control_server'
    SECRET_KEY_KEY = 'SECRET_KEY'

    def __init__(self):
        super(ControlServerConfiguration, self).__init__()

        self.flask_configuration = None
        self.control_server_configuration = None

        self.server_base_url = None

    def get_local_configuration_file_path(self):
        return (os.getenv(common_configuration.CONTROL_SERVER_CONFIG_ENVVAR_NAME) or
                common_configuration.DEFAULT_CONTROL_SERVER_CONFIG_PATH)

    def get_flask_configuration(self):
        was_updated = self.load_local_configuration()
        if was_updated or self.flask_configuration is None:
            flask_configuration = yaml.load(
                get_package_file_content('packy_agent.configuration.control_server',
                                         'flask.yaml'))
            deep_update(flask_configuration,
                        self.local_configuration.get(self.FLASK_CONFIGURATION_KEY) or {})

            secret_key_key = self.SECRET_KEY_KEY
            if not flask_configuration.get(secret_key_key):
                secret_key = generate_flask_secret_key()
                flask_configuration[secret_key_key] = secret_key
                self.local_configuration.setdefault(
                    self.FLASK_CONFIGURATION_KEY, {})[secret_key_key] = secret_key

                self.save_local_configuration()

            self.flask_configuration = flask_configuration

        return self.flask_configuration

    def get_control_server_configuration(self):
        was_updated = self.load_local_configuration()
        if was_updated or self.control_server_configuration is None:
            control_server_configuration = yaml.load(
                get_package_file_content('packy_agent.configuration.control_server',
                                         'custom.yaml'))
            deep_update(control_server_configuration,
                        (self.local_configuration.get(self.PACKY_CONFIGURATION_KEY) or {}).get(
                         self.CONTROL_SERVER_CONFIGURATION_KEY) or {})

            self.control_server_configuration = control_server_configuration
            self.override_server_base_url()

        return self.control_server_configuration

    def set_server_base_url_override(self, server_base_url):
        self.server_base_url = server_base_url

    def override_server_base_url(self, server_base_url=None):
        self.server_base_url = server_base_url or self.server_base_url

        self.control_server_configuration['server_base_url'] = (
            self.server_base_url or
            os.getenv(common_configuration.PACKY_SERVER_BASE_URL_ENVVAR_NAME) or
            # TODO(dmu) MEDIUM: Following line is for backward-compatibility only. Drop it once
            #                   all agents upgrade to v0.3.3 or higher
            os.getenv('DEFAULT_PACKY_SERVER_BASE_URL') or
            self.control_server_configuration['server_base_url'] or
            common_configuration.DEFAULT_SERVER_BASE_URL
        )

    def get_server_base_url(self):
        return self.get_control_server_configuration().get('server_base_url')

    def get_supervisor_configuration_file(self):
        return self.get_control_server_configuration().get('supervisor_configuration_file')

    def get_uwsgi_configuration_file(self):
        return self.get_control_server_configuration().get('uwsgi_configuration_file')

    def get_nginx_configuration_file(self):
        return self.get_control_server_configuration().get('nginx_configuration_file')

    def get_supervisor_log_file(self):
        return self.get_control_server_configuration().get('supervisor_log_file')

    def get_supervisor_log_directory(self):
        return os.path.dirname(self.get_supervisor_log_file())

    def get_supervisor_run_directory(self):
        return self.get_control_server_configuration().get('supervisor_run_directory')

    def get_control_server_stdout_log_file(self):
        return self.get_control_server_configuration().get('control_server_stdout_log_file')

    def get_control_server_stderr_log_file(self):
        return self.get_control_server_configuration().get('control_server_stderr_log_file')

    def get_agent_stdout_log_file(self):
        return self.get_control_server_configuration().get('agent_stdout_log_file')

    def get_agent_stderr_log_file(self):
        return self.get_control_server_configuration().get('agent_stderr_log_file')

    def get_watchdog_stdout_log_file(self):
        return self.get_control_server_configuration().get('watchdog_stdout_log_file')

    def get_watchdog_stderr_log_file(self):
        return self.get_control_server_configuration().get('watchdog_stderr_log_file')

    def get_uwsgi_pid_file(self):
        return self.get_control_server_configuration().get('uwsgi_pid_file')

    def get_packy_agent_pypi_repository(self):
        return self.get_control_server_configuration().get('packy_agent_pypi_repository')

    def is_reboot_enabled(self):
        return self.get_control_server_configuration().get('enable_reboot')

    def is_upgrade_enabled(self):
        return self.get_control_server_configuration().get('enable_upgrade')

    def is_network_configuration_enabled(self):
        return self.get_control_server_configuration().get('enable_network_configuration')

    def get_network_interface(self):
        return self.get_control_server_configuration().get('network_interface')

    def get_interfaces_config_path(self):
        return self.get_control_server_configuration().get('interfaces_config_path')

    def is_debug_mode(self):
        return (self.get_control_server_configuration().get('is_debug_mode') or
                super(ControlServerConfiguration, self).is_debug_mode())

    def get_control_server_port(self):
        return self.get_control_server_configuration().get('http_port')

    def set_control_server_port(self, port):
        self.load_local_configuration()
        self.local_configuration.setdefault('packy', {}).setdefault(
            'control_server', {})['http_port'] = port
        self.save_local_configuration()


configuration = ControlServerConfiguration()



from __future__ import absolute_import

import os
import select
import socket
import time
import logging
from contextlib import closing

from ryu.lib.packet.icmp import ICMP_ECHO_REPLY, ICMP_TIME_EXCEEDED
from ryu.lib.packet.ipv4 import ipv4
from ryu.lib.packet.packet import Packet

from packy_agent.modules.base.socket import (get_seq_no, TIME_SIZE_BYTES,
    send_icmp_echo_request as send_icmp_probe, decode_timestamp)
from packy_agent.modules.base.constants import (IP_PACKET_HEADER_SIZE_BYTES,
    ICMP_PACKET_HEADER_SIZE_BYTES)
from packy_agent.modules.traceroute.methods.base import guess_reply_hop_number


PACKET_HEADERS_SIZE_BYTES = IP_PACKET_HEADER_SIZE_BYTES + ICMP_PACKET_HEADER_SIZE_BYTES

logger = logging.getLogger(__name__)


def receive_icmp_reply_basic(raw_socket, id_, recieve_bytes=2048):
    ip_packet_data, (ip_address, _) = raw_socket.recvfrom(recieve_bytes)

    packet = Packet(ip_packet_data, parse_cls=ipv4)

    ip_packet = packet.protocols[0]
    icmp_packet = packet.protocols[1]
    logger.debug('Got ICMP packet type:{} code:{} from {}'.format(
        icmp_packet.type, icmp_packet.code, ip_address))
    if icmp_packet.type in (ICMP_TIME_EXCEEDED, ICMP_ECHO_REPLY):
        if icmp_packet.type == ICMP_TIME_EXCEEDED:
            inner_packet = Packet(icmp_packet.data.data, parse_cls=ipv4)
            effective_icmp_packet = inner_packet.protocols[1]
            is_destination_reached = False
        elif icmp_packet.type == ICMP_ECHO_REPLY:
            effective_icmp_packet = icmp_packet
            is_destination_reached = True
        else:
            assert False, 'Should never get here'
            raise ValueError('Unsupported ICMP packet type: {}'.format(icmp_packet.type))

        effective_id = effective_icmp_packet.data.id
        if effective_id != id_:
            return ip_address, ip_packet, icmp_packet, None, None, False

        effective_seq = effective_icmp_packet.data.seq
        effective_payload = effective_icmp_packet.data.data

        logger.debug(
            'ICMP packet from {} type:{} code:{} has effective id:{}, seq:{}, payload:{}'.format(
                ip_address, icmp_packet.type, icmp_packet.code, effective_id, effective_seq,
                effective_payload.encode('hex') if effective_payload else effective_payload))

        return (ip_address, ip_packet, icmp_packet, effective_icmp_packet, is_destination_reached,
                True)
    else:
        return ip_address, ip_packet, icmp_packet, None, None, False


def does_payload_match(payload, effective_payload, icmp_packet, ip_address):
    payload_len = len(payload)
    if effective_payload:
        effective_payload_len = len(effective_payload)
        if effective_payload_len != payload_len:
            logger.warning('Received payload length ({} bytes) differ from expected (sent) '
                           'length ({} bytes), but we are forgiving'.format(
                effective_payload_len, payload_len))
            min_len = min(effective_payload_len, payload_len)
            if effective_payload[:min_len] != payload[:min_len]:
                logger.debug('Discarded ICMP packet type:{} code:{} from {}: '
                             'payload prefix differs'.format(icmp_packet.type,
                                                             icmp_packet.code, ip_address))
                return False
        elif effective_payload != payload:
            logger.debug('Discarded ICMP packet type:{} code:{} from {}: '
                         'payload differs'.format(icmp_packet.type, icmp_packet.code,
                                                  ip_address))
            return False

    return True


def get_sent_time_from_payload(effective_payload):
    if effective_payload and len(effective_payload) >= TIME_SIZE_BYTES:
        return decode_timestamp(effective_payload[:TIME_SIZE_BYTES])


def receive_icmp_reply(raw_socket, id_, seq_no, payload, timeout, sent_time):
    payload_len = len(payload)

    time_left = timeout
    while time_left > 0:
        start_time = time.time()
        ready_sockets = select.select((raw_socket,), (), (), time_left)
        finish_time = time.time()  # because by this time the entire packet is already in OS buffer
        time_left -= (finish_time - start_time)

        if not ready_sockets[0]:  # timeout
            break

        (ip_address, ip_packet, icmp_packet, effective_icmp_packet, is_destination_reached,
         is_probe_response) = receive_icmp_reply_basic(raw_socket, id_, 1024 + payload_len)

        if not is_probe_response:
            logger.debug('Discarded ICMP packet type:{} code:{} from {}'.format(
                icmp_packet.type, icmp_packet.code, ip_address))
            continue

        effective_seq = effective_icmp_packet.data.seq
        effective_payload = effective_icmp_packet.data.data

        if effective_seq != seq_no:
            continue

        if not does_payload_match(payload, effective_payload, icmp_packet, ip_address):
            continue

        sent_time_from_payload = get_sent_time_from_payload(effective_payload)
        if sent_time_from_payload is not None:
            sent_time = sent_time_from_payload

        from packy_agent.modules.traceroute.base import ProbeResult
        return ProbeResult(
            hop_number=None,
            reply_hop_number=guess_reply_hop_number(ip_packet.ttl),
            probe_number=None,
            hop_ip_address=ip_address,
            rtt_seconds=finish_time - sent_time,
            is_destination_reached=is_destination_reached)


def trace_hop_with_icmp(destination_ip_address, timeout, ttl, packet_size_bytes=60):
    icmp_protocol = socket.getprotobyname('icmp')
    try:
        with closing(socket.socket(socket.AF_INET, socket.SOCK_RAW, icmp_protocol)) as raw_socket:
            raw_socket.setsockopt(socket.SOL_IP, socket.IP_TTL, ttl)

            id_ = os.getpid() & 0xFFFF
            seq_no = get_seq_no('icmp')
            sent_time, payload = send_icmp_probe(
                raw_socket, destination_ip_address, id_, seq_no,
                packet_size_bytes - PACKET_HEADERS_SIZE_BYTES)
            return receive_icmp_reply(raw_socket, id_, seq_no, payload, timeout,
                                      sent_time=sent_time)

    except socket.error as (errno, msg):
        if errno == 1:  # Operation not permitted
            msg += ' - Note that ICMP messages can only be sent from processes running as root.'
            raise socket.error(msg)

        raise



from __future__ import absolute_import

import raven

import logging
import packy_agent
from packy_agent.utils.misc import is_inside_docker_container
from packy_agent.configuration.agent.base import configuration

logger = logging.getLogger(__name__)


def get_raven_client(component_name):
    packy_agent_sentry_dsn = configuration.get_packy_agent_sentry_dsn()

    if packy_agent_sentry_dsn:
        agent_id = configuration.get_agent_id()
        name = '[{}, {}] {}'.format(
            # TODO(dmu) MEDIUM: Improve OS/platform reporting
            component_name, 'Docker' if is_inside_docker_container() else 'non-Docker',
            'Inactive agent' if agent_id is None else 'Agent ID: {}'.format(agent_id))

        return raven.Client(
            packy_agent_sentry_dsn,
            name=name,
            release=packy_agent.__version__,
            environment=configuration.get_server_base_url() or ''
        )
    else:
        logger.warning('Sentry DSN is not configured')



from setuptools import setup

from itertools import chain

version = __import__('packy_agent').__version__

# These packages are auto installed along with virtualenv creation
# pip==10.0.1
# setuptools==36.0.1
# wheel==0.29.0

# This dependency enumeration is required to maintain strict versions of libraries used
# More detailed explanation could be found here:
# https://dmugtasimov-tech.blogspot.ru/2016/12/strict-dependencies.html

PIP = ('pip==10.0.1',)  # TODO(dmu) LOW: Remove, not used in source code
PYTZ = ('pytz==2017.2',)
CERTIFI = ('certifi==2017.4.17',)
PYPARSING = ('pyparsing==2.2.0',)
PYYAML = ('PyYAML==3.12',)
ATOMICWRITES = ('atomicwrites==1.1.5',)
CACHETOOLS = ('cachetools==2.0.0',)
PYTHON_DATEUTIL = ('python-dateutil==2.6.1',)

CELERY = ('celery==4.1.0', 'amqp==2.2.2', 'billiard==3.5.0.3', 'kombu==4.1.0', 'vine==1.1.4') + PYTZ
REQUESTS = ('requests==2.18.4', 'chardet==3.0.4', 'idna==2.6', 'urllib3==1.22') + CERTIFI
ITSDANGEROUS = ('itsdangerous==0.24',)
SPEEDTEST_CLI = ('speedtest-cli==1.0.7',)
SUPERVISOR = ('supervisor==3.3.3', 'meld3==1.0.2')
JINJA = ('Jinja2==2.9.6',)
NETIFACES = ('netifaces==0.10.6',)
TAILER = ('tailer==0.4.1',)
UWSGI = ('uWSGI==2.0.15',)
UPTIME = ('uptime==3.0.1',)
SCHEMATICS = ('schematics==2.0.1',)
PYCURL = ('pycurl==7.43.0.2',)
RETRY = ('retry==0.9.2',)
PSUTIL = ('psutil==5.4.0',)
RYU = ('ryu==4.23',)
FILELOCK = ('filelock==3.0.4',)
ANSIBLE = ('ansible==2.5.1',)

FLASK = ('Flask==0.12.2', 'click==6.7', 'MarkupSafe==1.0',
         'Werkzeug==0.12.2') + ITSDANGEROUS + JINJA
FLASK_WTF = ('Flask-WTF==0.14.2',) + FLASK
RAVEN = ('raven[flask]==6.9.0',)
DISTRO = ('distro==1.3.0',)

install_requires = list(set(chain(
    CELERY, REQUESTS, PYPARSING, PYYAML, FLASK, FLASK_WTF, SPEEDTEST_CLI,
    ATOMICWRITES, CACHETOOLS, SUPERVISOR, JINJA, PIP, NETIFACES, TAILER, UWSGI, UPTIME,
    PYTHON_DATEUTIL, SCHEMATICS, RETRY, PSUTIL, RYU, FILELOCK, ANSIBLE, RAVEN, DISTRO
)))

setup(
    name='packy-agent',
    version=version,
    description='Packy Agent',
    packages=['packy_agent'],
    zip_safe=False,
    install_requires=install_requires,
    extras_require={
        'post': list(PYCURL),
        'dev': ['behave==1.2.6', 'readchar==0.7', 'six==1.11.0']
    },
    include_package_data=True,
    entry_points={
        'console_scripts': [
            # TODO(dmu) LOW: Rename `control-server` -> `packy-control-server`
            'control-server = packy_agent.control_server.run:entry',
            'packy-watchdog = packy_agent.watchdog.run:entry',
            'packy-welcome = packy_agent.cli.welcome:entry',
            'packy-agent-activate = packy_agent.cli.activate:entry',
        ]
    }
)



from __future__ import absolute_import

import logging
import subprocess

logger = logging.getLogger(__name__)


def run_command(command, raise_exception=False):
    logger.info('Running shell command: %s', command)
    try:
        output = subprocess.check_output(command, shell=True, stderr=subprocess.STDOUT).strip()
    except subprocess.CalledProcessError as ex:
        logger.error('Shell command failed: %s: %s', command, ex.output.strip())
        if raise_exception:
            raise
    else:
        logger.info('Shell command succeeded: %s: %s', command, output.strip())
        return output


def install_ubuntu_packages(packages, one_command=True):
    run_command('apt-get update')
    if one_command:
        run_command('apt-get install -y {}'.format(' '.join(packages)))
    else:
        for package in packages:
            run_command('apt-get install -y {}'.format(package))


def install_alpine_packages(packages, one_command=True):
    run_command('apk update')
    if one_command:
        run_command('apk add {}'.format(' '.join(packages)))
    else:
        for package in packages:
            run_command('apk add {}'.format(package))



from datetime import datetime

from schematics import types

from packy_agent.utils.schematics import CustomModel


class BaseMeasurement(CustomModel):
    dt = types.DateTimeType(default=datetime.utcnow)


class HTTPModuleMeasurement(BaseMeasurement):
    ts = types.IntType(required=True)
    target = types.StringType(required=True)

    is_success = types.BooleanType(default=True)
    http_status_code = types.IntType()

    namelookup_ms = types.FloatType()
    total_ms = types.FloatType()

    certificate_expiration_dt = types.DateTimeType()



from __future__ import absolute_import

import logging
import random
import os
import os.path
import netifaces
import platform
import distutils.spawn
import datetime
import pytz
import dateutil.parser

from atomicwrites import AtomicWriter, atomic_write as atomic_write_original

import packy_agent
from packy_agent.clients.packy_server import PackyServerClient
from packy_agent.exceptions import AuthenticationError

UBUNTU_16_04_XENIAL = ('Ubuntu', '16.04', 'xenial')
LINUXMINT_18_1_SERENA = ('LinuxMint', '18.1', 'serena')

EPOCH_START_DATETIME = datetime.datetime(1970, 1, 1, tzinfo=pytz.utc)
UPGRADE_IN_PROGRESS_FILE_LOCK = '/tmp/packy-upgrade-in-progress.lock'


logger = logging.getLogger(__name__)


def get_executable_path(executable, self_failover=True):
    path = distutils.spawn.find_executable(executable)
    if not path and self_failover:
        path = executable

    return path


def get_operating_system():
    try:
        return platform.linux_distribution()
    except Exception:
        # in case someone run it not on linux
        return None


def get_interfaces():
    return netifaces.interfaces()


def is_inside_docker_container():
    return os.path.isfile('/.dockerenv')


class CustomAtomicWriter(AtomicWriter):
    def commit(self, f):
        path = self._path
        if os.path.isfile(path):
            filestat = os.stat(path)
        else:
            filestat = None

        super(CustomAtomicWriter, self).commit(f)

        if filestat is not None:
            try:
                os.chmod(path, filestat.st_mode)
                os.chown(path, filestat.st_uid, filestat.st_gid)
            except Exception:
                logger.warning('Failed to restore permissions or ownership for file: %s', path)


def atomic_write(path, writer_cls=CustomAtomicWriter, **cls_kwargs):
    return atomic_write_original(path=path, writer_cls=writer_cls, **cls_kwargs)


# TODO(dmu) MEDIUM: Use `packy_agent.utils.shell.run_command` instead?
# TODO(dmu) MEDIUM: Provide logging to file for running shell commands
def run_shell_command_async(command, delay_seconds=None, dev_null=False):
    if delay_seconds is None:
        command = '({}) &'.format(command)
    else:
        command = '(sleep {}; {}) &'.format(delay_seconds, command)

    if dev_null:
        command = '({}) > /dev/null 2>&1'.format(command)

    logger.debug('Started async shell command: %s', command)
    os.system(command)
    logger.debug('Finished async shell command: %s', command)


def generate_random_string(size_bytes):
    return (('{:0' + str(size_bytes) + 'x}').format(random.getrandbits(size_bytes * 4))
            if size_bytes else '')


def get_iso_format_utcnow():
    return datetime.datetime.utcnow().isoformat() + 'Z'


def to_epoch(dt):
    return (dt - EPOCH_START_DATETIME).total_seconds()


def iso_to_datetime(dt_string):
    return dateutil.parser.parse(dt_string)


def dump_version(filename, variable_name):
    with open(filename, 'w') as f:
        f.write('{}={}\n'.format(variable_name, packy_agent.__version__))


def get_inactive_agents(email, password):
    from packy_agent.configuration.control_server.base import configuration
    client = PackyServerClient(configuration.get_server_base_url())
    response = client.get_inactive_agents(email, password)

    if response.status_code == 401:
        raise AuthenticationError('Invalid credentials for agent activation')

    response.raise_for_status()
    return {agent['id']: agent['name'] for agent in response.json()}


def is_upgrade_in_progress():
    return os.path.isfile(UPGRADE_IN_PROGRESS_FILE_LOCK)


def remove_upgrade_in_progress_lock():
    try:
        os.remove(UPGRADE_IN_PROGRESS_FILE_LOCK)
    except OSError:
        pass



import logging

import flask
from flask.views import MethodView

from packy_agent.control_server.forms.activate import ActivateForm
from packy_agent.control_server.forms.activate_extra import ActivateExtraForm
from packy_agent.control_server.views.base import smart_redirect
from packy_agent.exceptions import AuthenticationError, ValidationError
from packy_agent.managers.install_and_upgrade import install_and_upgrade_manager
from packy_agent.utils.auth import set_authentication_cookie, is_activated
from packy_agent.utils.misc import get_inactive_agents

logger = logging.getLogger(__name__)


class ActivateView(MethodView):

    def get(self):
        if is_activated():
            return flask.redirect(flask.url_for('index'))

        return flask.render_template('activate.html', form=ActivateForm())

    def post(self):
        if is_activated():
            return flask.redirect(flask.url_for('index'))

        form = ActivateForm()
        if form.validate():
            email = form.email.data
            password = form.password.data

            try:
                agent_id = None
                agents = get_inactive_agents(email, password)
                if agents:
                    form = ActivateExtraForm(email=email, password=password)
                    agent_choices = tuple((str(id_), name) for id_, name in agents.iteritems())
                    form.agent.choices += agent_choices

                    if form.validate():
                        if form.agent.data != 'new':
                            agent_id = int(form.agent.data)
                            if agent_id not in agents:
                                flask.flash('Invalid agent')
                                return flask.redirect(flask.url_for('activation_failure'))
                    elif form.extra.data == 'yes':
                        return flask.render_template('activate_extra.html', form=form)
                    else:
                        form = ActivateExtraForm(email=email, password=password)
                        form.agent.choices += agent_choices
                        return flask.render_template('activate_extra.html', form=form)

                install_and_upgrade_manager.activate(email, password, agent_id=agent_id)

            except AuthenticationError:
                flask.flash('Not authenticated (invalid credentials)')
                return flask.redirect(flask.url_for('activation_failure'))
            except ValidationError as ex:
                flask.flash(ex.message)
                return flask.redirect(flask.url_for('activation_failure'))
            except Exception:
                logger.exception('Error during activation')
                flask.flash('Error during activation')
                return flask.redirect(flask.url_for('activation_failure'))

            set_authentication_cookie()
            return smart_redirect('success', 'index', button_text='Continue')

        return flask.render_template('activate.html', form=form)


class ActivationFailureView(MethodView):

    def get(self):
        return flask.render_template('activation_failure.html')



import logging
import socket
import time

from celery import shared_task
from retry.api import retry_call

from packy_agent.clients.packy_server import packy_server_client_factory
from packy_agent.configuration.agent.base import configuration
from packy_agent.modules.ping.base import ping_once
from packy_agent.tasks.modules.base import run_exclusive
from packy_agent.utils.misc import get_iso_format_utcnow

logger = logging.getLogger(__name__)


def ping(host):
    module_config = configuration.get_ping_module_configuration()

    number_of_pings = module_config['number_of_pings']
    packet_size = module_config['packet_size']
    interval_ms = module_config.get('interval_ms', 0)
    start_time = get_iso_format_utcnow()
    pings = []

    try:
        destination_ip_address = socket.gethostbyname(host)
    except socket.gaierror:
        logger.warning('Could not resolve {} to IP address'.format(host))
    else:
        for _ in xrange(number_of_pings):
            start = time.time()
            try:
                rtt_seconds = ping_once(destination_ip_address, 10, packet_size_bytes=packet_size)
                if rtt_seconds is None:
                    continue
            except Exception:
                logger.warning('Error during ping', exc_info=True)
                continue

            pings.append(round(rtt_seconds * 1000, 2))

            time_left = interval_ms / 1000. - (time.time() - start)
            if time_left > 0:
                logger.debug('Waiting {0:.03f} seconds for next ping'.format(time_left))
                time.sleep(time_left)

    # TODO(dmu) HIGH: Producing of result structure should be moved to PackyServerClient level
    return {
        'target': host,
        'packet_size': packet_size,
        'n_pings': number_of_pings,
        'time': start_time,
        'pings': ','.join(map(str, pings))
    }


def ping_all_impl():
    hosts = configuration.get_ping_module_configuration().get('hosts')
    if not hosts:
        logger.warning('Ping hosts are not configured')
        # TODO(dmu) HIGH: Should I mark task as failed?
        return

    client = packy_server_client_factory.get_client_by_configuration(configuration)
    # See comment for speedtest
    for host in hosts:
        result = ping(host)
        retry_call(client.submit_ping_measurement, (result,),
                   tries=configuration.get_submit_retries())


@shared_task
def ping_all():
    run_exclusive(ping_all_impl)



from __future__ import absolute_import

import functools
from packy_agent.configuration.agent.base import configuration as agent_configuration
from flask import session, redirect, request, url_for


def is_activated():
    return agent_configuration.is_activated()


def is_authenticated():
    agent_id = session.get('agent_id')
    if agent_id is not None:
        return agent_configuration.get_agent_id() == agent_id
    return False


def set_authentication_cookie():
    agent_id = agent_configuration.get_agent_id()
    if agent_id is not None:
        session['agent_id'] = agent_id


def login():
    set_authentication_cookie()


def logout():
    session.pop('agent_id', None)


def activation_and_authentication_required(func_or_method):
    @functools.wraps(func_or_method)
    def wrapper(*args, **kwargs):
        if not is_activated():
            return redirect(url_for('activate'))
        elif not is_authenticated():
            return redirect(url_for('login') + '?next=' + request.path)
        else:
            return func_or_method(*args, **kwargs)

    return wrapper


def activation_required(func_or_method):
    @functools.wraps(func_or_method)
    def wrapper(*args, **kwargs):
        if not is_activated():
            return redirect(url_for('activate'))
        else:
            return func_or_method(*args, **kwargs)

    return wrapper
